from pycparser import c_ast

from six import iteritems, iterkeys

from . import ASTVisitor
from .. import *  # noqa
from ..types import *  # noqa

class CodegenVisitor(ASTVisitor):
  priority = 1000

  def __init__(self, *args, **kwargs):
    super(CodegenVisitor, self).__init__(*args, **kwargs)

    self.types = {
      'void':          VoidType(self),
      'int':           IntType(self),
      'unsigned int':  UnsignedIntType(self),
      'char':          CharType(self),
      'unsigned char': UnsignedCharType(self)
    }
    self.types['unsigned'] = self.types['unsigned int']

    for desc in list(iterkeys(self.types)):
      self.types[desc + ' *'] = PointerType(self.types[desc], self)

    self.functions = []
    self.global_scope = Scope(self)

    self.FN       = None
    self.BLOCK    = None
    self.SCOPE    = self.global_scope
    self.EMIT     = None
    self.GET_REG  = None
    self.REGS     = None

    self.blocks = []
    self.prolog_blocks = []
    self.epilog_blocks = []

    self.label_index = 0
    self.literal_label_index = 0
    self.string_literals = {}

    self.break_stack = []
    self.continue_stack = []

    self.make_current(self.block(name = self.get_new_label(name = 'global')))

  def block(self, stage = None, *args, **kwargs):
    block = Block(*args, **kwargs)

    if stage == 'prolog':
      self.prolog_blocks.append(block)

    elif stage == 'epilog':
      self.epilog_blocks.append(block)

    else:
      self.blocks.append(block)

    return block

  def make_current(self, block):
    self.DEBUG(self.log_prefix + 'new current block: %s', block)

    self.BLOCK = block
    self.EMIT = block.emit

  def reset_scope(self):
    while self.SCOPE != self.global_scope:
      self.pop_scope()

  def push_scope(self):
    self.SCOPE = Scope(self, parent = self.SCOPE)

    self.DEBUG(self.log_prefix + 'scope: pushing new scope: scope=%s, parent=%s', self.SCOPE, self.SCOPE.parent)

    return self.SCOPE

  def pop_scope(self):
    self.DEBUG(self.log_prefix + 'scope: popping scope: scope=%s, parent=%s', self.SCOPE, self.SCOPE.parent)

    self.SCOPE = self.SCOPE.parent

  def get_new_label(self, name = None):
    self.label_index += 1
    return '.L%i' % self.label_index if name is None else '.L%i_%s' % (self.label_index, name)

  def get_new_literal_label(self):
    self.literal_label_index += 1
    return '.LC%i' % self.literal_label_index

  def get_new_local_storage(self, size):
    assert self.FN is not None

    if size <= 4:
      self.FN.fp_offset -= 4

    else:
      self.WARN('Unhandled get_new_local_storage branch: size=%s', size)

    return StackSlotStorage(None, self.FN.fp_offset)

  def emit_string_literals(self):
    B = self.block(stage = 'prolog', comment = 'string literals block')

    B.emit(Directive('.include "ducky.asm"'))
    B.emit(Directive('.section .rodata'))

    for label, s in iteritems(self.string_literals):
      B.emit(Directive('.type %s, string' % label))
      B.emit(Directive('.string %s' % s))

  def emit_trampoline(self):
    pass

  def emit_prolog(self):
    B = self.block(stage = 'prolog', comment = 'global pre-prolog')
    B.emit(Directive('.data'))
    B.emit(Directive('.type __global_stack, space'))
    B.emit(Directive('.space 256'))
    B.emit(Directive('.text'))

    B = self.block(stage = 'prolog', name = '_start', comment = 'global prolog')
    B.emit(LA('sp', '&__global_stack'))
    B.emit(ADD('sp', 256))
    B.emit(MOV('fp', 'sp'))
    B.emit(CALL('&main'))
    B.emit(HLT('r0'))
    B.emit(Directive('.global _start'))

    self.emit_string_literals()

  def emit_epilog(self):
    self.emit_trampoline()

  def materialize(self):
    # Emit translation unit prolog and epilog
    self.emit_prolog()
    self.emit_epilog()

    # Now materialize block tree
    code = []

    for block in self.prolog_blocks:
      block.materialize(code)
      code.append('')

    for block in self.blocks:
      block.materialize(code)
      code.append('')

    for fn in self.functions:
      code += fn.materialize()
      code.append('')

    for block in self.epilog_blocks:
      block.materialize(code)
      code.append('')

    return '\n'.join(code)

  def visit(self, node, **kwargs):
    method = 'visit_' + node.__class__.__name__
    visitor = getattr(self, method, self.generic_visit)

    if visitor == self.generic_visit:
      self.WARN(self.log_prefix + 'Unhandled node class: %s', node.__class__.__name__)

    return visitor(node, **kwargs)

  def generic_visit(self, node, **kwargs):
    old_parent = self.node_parent
    self.node_parent = node

    ret = []
    for c_name, c in node.children():
      ret.append(self.visit(c, **kwargs))

    self.node_parent = old_parent

    return ret

  def visit_constant_value(self, node):
    self.DEBUG(self.log_prefix + 'visit_constant_value: %s', dump_node(node))

    if node.type == 'string':
      label = self.get_new_literal_label()
      self.string_literals[label] = node.value

      return RValueExpression(value = StringConstantValue('&%s' % label), type = CType.get_from_desc(self, 'char *'))

    if node.type == 'int':
      return RValueExpression(value = ConstantValue(node.value), type = CType.get_from_desc(self, 'int'))

    if node.type == 'char':
      special_chars = [r'\0']

      c = node.value[1:-1]
      if c in special_chars:
        return RValueExpression(value = ConstantValue(str(special_chars.index(c))), type = CType.get_from_desc(self, 'char'))

      return RValueExpression(value = ConstantValue(str(ord(c))), type = CType.get_from_desc(self, 'char'))

    raise CompilerError(node.coord, 'Unhandled branch: Constant, type=%s' % node.type)

  def visit_expr(self, node, preferred = None, keep = None):
    self.DEBUG(self.log_prefix + 'visit_expr: node=%s, preferred=%s, keep=%s', dump_node(node), preferred, keep)
    self.DOWN()

    E = self.visit(node, preferred = preferred, keep = keep)
    self.DEBUG(self.log_prefix + 'visit_expr: expr=%s', E)

    self.UP()
    return E

  def process_cond(self, node, iftrue_label = None, iffalse_label = None):
    CE = self.visit(node)
    self.DEBUG(self.log_prefix + 'cond=%s', CE)

    if CE.value is None:
      # check flags according to our cond

      if isinstance(node, c_ast.BinaryOp):
        if iftrue_label is not None:
          if node.op == '==':
            self.EMIT(BE('&' + iftrue_label))

          elif node.op == '!=':
            self.EMIT(BNE('&' + iftrue_label))

          elif node.op == '<=':
            self.EMIT(BLE('&' + iftrue_label))

          elif node.op == '>=':
            self.EMIT(BGE('&' + iftrue_label))

          else:
            self.WARN(self.log_prefix + 'Condition handling not implemented: BinaryOp, op=%s', node.op)

        if iffalse_label is not None:
          if node.op == '==':
            self.EMIT(BNE('&' + iffalse_label))

          elif node.op == '!=':
            self.EMIT(BE('&' + iffalse_label))

          elif node.op == '<=':
            self.EMIT(BG('&' + iffalse_label))

          elif node.op == '>=':
            self.EMIT(BL('&' + iffalse_label))

          else:
            self.WARN(self.log_prefix + 'Condition handling not implemented: BinaryOp, op=%s', node.op)

      else:
        self.WARN(self.log_prefix + 'Condition handling not implemented: cond=%s', dump_node(node))

  def _perform_implicit_conversion(self, loc, E, T):
    self.DEBUG(self.log_prefix + '_perform_implicit_conversion: E=%s, T=%s', E, T)
    self.DOWN()

    assert E.is_rvalue()

    # Any integer type rvalue can be converted to any other integer type, with
    # possible loss of precision.
    if isinstance(E.type, IntType):
      if isinstance(T, UnsignedIntType):
        value = E.value

        if isinstance(value, ConstantValue):
          if int(value.value) < 0:
            value = 2 ^ 16 + int(value.value)

          else:
            value = value.value

          self.UP()
          return RValueExpression(value = ConstantValue(str(value)), type = T), E

        elif isinstance(value, RegisterValue):
          self.UP()
          return RValueExpression(value = value, type = T), E

        else:
          self.WARN(self.log_prefix + 'Unhandled conversion branch')

    if isinstance(E.type, UnsignedIntType):
      if isinstance(T, IntType):
        value = E.value

        if isinstance(value, ConstantValue):
          if int(value.value) >= 32768:
            self.WARN(self.log_prefix + 'Implicit "%s" -> "%s" conversion - value "%s" cannot fit into destination', E.type, T, value.value)

          self.UP()
          return RValueExpression(value = ConstantValue(str(value.value)), type = T), E

        elif isinstance(value, RegisterValue):
          self.UP()
          return RValueExpression(value = value, type = T), E

        else:
          self.WARN(self.log_prefix + 'Unhandled conversion branch')

    if isinstance(E.type, PointerType):
      if isinstance(T, PointerType):
        # void * can be casted to any pointer, and any pointer can be casted to void *
        if isinstance(E.type.ptr_to_type, VoidType) or isinstance(T.ptr_to_type, VoidType):
          self.UP()
          return RValueExpression(value = E.value, type = T), E

    raise IncompatibleTypesError(loc, T, E.type)

  #
  # Real visitors
  #

  def visit_ArrayRef(self, node, preferred = None, keep = None):
    comment = dump_node(node)

    self.INFO(self.log_prefix + comment)
    self.DOWN()
    self.EMIT(Comment(comment))

    AE = self.visit_expr(node.name)
    self.DEBUG(self.log_prefix + 'array=%s', AE)

    SE = self.visit_expr(node.subscript)
    self.DEBUG(self.log_prefix + 'subscript=%s', SE)

    if not SE.is_rvalue():
      SE, old_SE = SE.to_rvalue(self)
      self.DEBUG(self.log_prefix + 'subscript=%s', SE)

    reg = self.GET_REG(keep = [array_reg, subscript_reg])
    self.EMIT(MOV(reg.name, subscript))
    self.EMIT(MUL(reg.name, len(array_symbol.type.ptr_to_type)))
    self.EMIT(ADD(reg.name, array))

    self.PUT_REG(array_reg)
    self.PUT_REG(subscript_reg)

    self.EMIT(Comment('/ ' + comment))
    self.UP()
    return reg.name, reg, array_symbol

  def visit_Assignment(self, node, preferred = None, keep = None):
    comment = dump_node(node)

    self.INFO(self.log_prefix + comment)
    self.DOWN()
    self.EMIT(Comment(comment))

    LE = self.visit_expr(node.lvalue)
    RE = self.visit_expr(node.rvalue)

    self.DEBUG(self.log_prefix + 'left=%s', LE)
    self.DEBUG(self.log_prefix + 'right=%s', RE)

    if not LE.is_mlvalue():
      raise CompilerError(node.coord, 'Unable to assign to a read-only variable')

    if RE.is_lvalue():
      # Implicit lvalue->rvalue conversion must take a place here
      RE, old_RE = RE.to_rvalue(self)
      self.DEBUG(self.log_prefix + 'right=%s', RE)

    self.DEBUG(self.log_prefix + 'value "%s" to storage "%s", %i bytes', RE.type, LE.type, len(LE.type.ptr_to_type))

    if LE.type.ptr_to_type != RE.type:
      RE, old_RE = self._perform_implicit_conversion(node.coord, RE, LE.type.ptr_to_type)
      self.DEBUG(self.log_prefix + 'right=%s', RE)

    if LE.value.can_register_backed():
      if not LE.value.is_register_backed():
        self.EMIT(Comment('LE can be backed by register, no register acquired yet'))

        if isinstance(RE.value, RegisterValue) and RE.value.register.storage is None:
          LE.value.storage.acquire_register(RE.value.register)
          LE.value.backing_register().dirty = True

        else:
          R = self.GET_REG()
          LE.value.storage.acquire_register(R)
          self.EMIT(MOV(LE.value.backing_register().name, RE.value.name))

      else:
        self.EMIT(MOV(LE.value.backing_register().name, RE.value.name))

      LE.value.backing_register().dirty = True

    else:
      if len(LE.type.ptr_to_type) == 1:
        self.EMIT(STB(LE.value.name, RE.value.name))

      elif len(LE.type.ptr_to_type) == 2:
        self.EMIT(STW(LE.value.name, RE.value.name))

      else:
        self.WARN(self.log_prefix + 'Unhandled assignment size branch')

    if isinstance(LE.value, RegisterValue):
      LE.value.register.put()

    self.EMIT(Comment('/ ' + comment))
    self.UP()
    return RE

  def visit_BinaryOp(self, node, preferred = None, keep = None):
    comment = dump_node(node)

    self.INFO(self.log_prefix + '%s, preferred=%s, keep=%s', dump_node(node), preferred, keep)
    self.DOWN()
    self.EMIT(Comment(dump_node(node)))

    self.EMIT(Comment('left: ' + dump_node(node.left)))
    self.EMIT(Comment('right: ' + dump_node(node.right)))

    LE = self.visit_expr(node.left)
    RE = self.visit_expr(node.right)

    self.DEBUG(self.log_prefix + 'left=%s', LE)
    self.DEBUG(self.log_prefix + 'right=%s', RE)

    if not LE.is_rvalue():
      LE, old_LE = LE.to_rvalue(self)
      self.DEBUG(self.log_prefix + 'left=%s', LE)

    if not RE.is_rvalue():
      RE, old_RE = RE.to_rvalue(self)
      self.DEBUG(self.log_prefix + 'right=%s', RE)

    keep = keep or []

    def __binop_arith(inst):
      E = RValueExpression(value = RegisterValue(self.GET_REG(preferred = preferred, keep = keep + [LE.value, RE.value])), type = LE.type)

      if isinstance(LE.value, RegisterValue):
        self.EMIT(MOV(E.value.name, LE.value.name))

      elif isinstance(LE.value, StringConstantValue):
        self.EMIT(LA(E.value.name, LE.value.name))

      elif isinstance(LE.value, ConstantValue):
        self.EMIT(LI(E.value.name, LE.value.name))

      else:
        self.WARN(self.log_prefix + 'Unhandled binop arith branch')

      self.EMIT(inst(E.value.name, RE.value.name))

      return E

    if LE.type != RE.type:
      RE, old_RE = self._perform_implicit_conversion(node.coord, RE, LE.type)
      self.DEBUG(self.log_prefix + 'right=%s', RE)

    R = None

    if node.op == '+':
      R = __binop_arith(ADD)

    elif node.op == '-':
      R = __binop_arith(SUB)

    elif node.op == '*':
      ret, ret_reg = __binop_arith(MUL)

    elif node.op == '&':
      R = __binop_arith(AND)

    elif node.op == '||':
      R = __binop_arith(OR)

    elif node.op in ('!=', '>=', '<=', '==', '>', '<'):
      R = RValueExpression(type = IntType(self))

      self.EMIT(CMP(LE.value.name, RE.value.name))

    else:
      self.WARN(self.log_prefix + 'Unhandled binary op: op=%s', node.op)

    if isinstance(LE.value, RegisterValue):
      LE.value.register.put()
    if isinstance(RE.value, RegisterValue):
      RE.value.register.put()

    self.EMIT(Comment('/ ' + comment))
    self.UP()
    return R

  def visit_Cast(self, node, **kwargs):
    comment = dump_node(node)

    self.INFO(self.log_prefix + comment)
    self.DOWN()
    self.EMIT(Comment(comment))

    E = self.visit_expr(node.expr, **kwargs)
    self.DEBUG(self.log_prefix + 'expr=%s', E)

    if E.is_lvalue():
      E, old_E = E.to_rvalue(self)
      self.DEBUG(self.log_prefix + 'expr=%s', E)

    self.DEBUG(self.log_prefix + 'typename: %s', dump_node(node.to_type))
    T = CType.get_from_decl(self, node.to_type.type)
    self.DEBUG(self.log_prefix + 'type: %s', T)

    R = RValueExpression(value = RegisterValue(self.GET_REG(**kwargs)), type = T)

    self.EMIT(Comment('casting %s to %s' % (E, T)))
    self.EMIT(MOV(R.value.name, E.value.name))

    self.EMIT(Comment('/ ' + comment))
    self.UP()
    return R

  def visit_Compound(self, node, create_scope = True):
    self.INFO(self.log_prefix + dump_node(node))
    self.DOWN()

    self.EMIT(Comment(dump_node(node)))

    if create_scope:
      self.push_scope()

    ret = self.generic_visit(node)
    self.DEBUG(self.log_prefix + 'compound output: %s', ret)

    for r in ret:
      if r is None or not isinstance(r.value, RegisterValue):
        continue

      r.value.register.put()

    self.pop_scope()
    self.UP()

  def visit_Constant(self, node, preferred = None, keep = None):
    comment = dump_node(node)

    self.DEBUG(self.log_prefix + '%s: preferred=%s, keep=%s', comment, preferred, keep)
    self.DOWN()
    self.EMIT(Comment(comment))

    E = self.visit_constant_value(node)

    self.EMIT(Comment('/ ' + comment))
    self.UP()
    return E

  def visit_Decl(self, node):
    comment = dump_node(node)

    self.INFO(self.log_prefix + comment)
    self.DOWN()
    self.EMIT(Comment(comment))

    if node.name is None:  # declaring a type, structure probably... we'll see
      T = CType.create_from_decl(self, node)
      self.types[repr(T)] = T

      # self.EMIT(Comment('/ ' + comment))
      self.UP()
      return T

    symbol = self.SCOPE.add(node.coord, Symbol(self, node.name, CType.create_from_decl(self, node.type), extern = ('extern' in node.storage), const = ('const' in node.quals)))

    RE_class = MLValueExpression if 'const' in node.quals else LValueExpression
    RE = RE_class(type = PointerType(symbol.type, self))

    # get storage for new symbol
    if self.FN is None:
      storage = MemorySlotStorage(symbol, node.name)
      RE.value = MemorySlotValue(storage)

    else:
      storage = self.get_new_local_storage(len(RE.type.ptr_to_type))
      RE.value = StackSlotValue(storage)

    storage.symbol = symbol
    symbol.storage = storage

    self.DEBUG(self.log_prefix + 'storage: size=%i, slot=%s', len(RE.type.ptr_to_type), storage)

    self.EMIT(Comment('storage: %s => size=%i, slot=%s' % (node.name, len(RE.type.ptr_to_type), RE.value.name)))

    self.DEBUG(self.log_prefix + 'RE=%s', RE)

    if node.init is not None:
      self.DEBUG(self.log_prefix + 'has init, process it')
      self.DOWN()
      self.EMIT(Comment('init of %s' % comment))

      IE = self.visit(node.init)
      self.DEBUG(self.log_prefix + 'init=%s', IE)

      if isinstance(IE.value, RegisterValue):
        self.WARN(self.log_prefix + 'Unhandled init branch')

      else:
        R = self.GET_REG()
        R.dirty = True
        symbol.storage.acquire_register(R)

        self.EMIT(LI(R.name, IE.value.name))
        # self.EMIT(STW(RE.value.name, R.name))

      self.EMIT(Comment('/ init of %s' % comment))
      self.UP()

    self.DEBUG(self.log_prefix + 'declared symbol: %s', symbol)
    self.EMIT(Comment('declared symbol: %s' % symbol))

    # self.EMIT(Comment('/ ' + comment))
    self.UP()
    return RE

  def visit_ExprList(self, node):
    self.INFO(self.log_prefix + dump_node(node))
    self.DOWN()

    for expr in node.exprs:
      RE = self.visit_expr(expr)

    self.UP()
    return RE

  def visit_FileAST(self, node):
    self.generic_visit(node)

  def visit_For(self, node):
    comment = dump_node(node)

    self.INFO(self.log_prefix + comment)
    self.DOWN()
    self.EMIT(Comment(comment))

    init_block = self.FN.block(name = self.get_new_label(name = 'for_init'))
    head_block = self.FN.block(name = self.get_new_label(name = 'for_head'))
    body_block = self.FN.block(name = self.get_new_label(name = 'for_body'))
    past_block = self.FN.block(name = self.get_new_label(name = 'for_past'))

    self.break_stack.append(past_block)
    self.continue_stack.append(head_block)

    self.BLOCK.connect(init_block)
    init_block.connect(head_block)
    head_block.connect(body_block)
    head_block.connect(past_block)
    body_block.connect(head_block)
    body_block.connect(past_block)

    self.make_current(init_block)

    IE = self.visit(node.init)
    self.DEBUG(self.log_prefix + '  init=%s', IE)

    self.make_current(head_block)

    self.process_cond(node.cond, iftrue_label = body_block.names[0], iffalse_label = past_block.names[0])

    self.make_current(body_block)
    self.visit(node.stmt)

    self.visit(node.next)
    self.EMIT(J('&' + head_block.names[0]))

    self.make_current(past_block)

    self.break_stack.pop()
    self.continue_stack.pop()

    self.UP()

  def visit_FuncCall(self, node, preferred = None, keep = None):
    comment = dump_node(node)

    self.INFO(self.log_prefix + comment)
    self.DOWN()
    self.EMIT(Comment(comment))

    self.INFO(self.log_prefix + 'Name: %s', dump_node(node.name))
    self.INFO(self.log_prefix + 'Args: %s', dump_node(node.args))

    if isinstance(node.name, c_ast.ID) and node.name.name == 'asm' and node.args is not None and isinstance(node.args.exprs[0], c_ast.Constant) and node.args.exprs[0].type == 'string':
      for line in node.args.exprs[0].value[1:-1].split('\n'):
        self.EMIT(InlineAsm(line))

      self.UP()
      return None

    FE = self.visit(node.name)
    if FE.value is None:
      raise SymbolUndefined(node.coord, dump_node(node.name))

    FE_symbol = FE.value.storage.symbol

    self.DEBUG(self.log_prefix + 'fn=%s', FE)

    if node.args is not None:
      args = []
      restore_regs = []

      for i, arg in enumerate(node.args.exprs):
        arg_comment = 'arg #%i: %s' % (i, dump_node(arg))

        self.DEBUG(self.log_prefix + arg_comment)
        self.DOWN()
        self.EMIT(Comment(arg_comment))

        R = self.FN.registers.all_regs[i + 1]

        self.EMIT(Comment('free register'))
        if R.storage is None:
          restore_regs.append((R, None))
          self.EMIT(PUSH(R.name))

        else:
          restore_regs.append((R, R.storage))
          R.free()

        self.DEBUG(self.log_prefix + 'place into %s', R)

        A = self.visit(arg, preferred = R.index)

        self.DEBUG(self.log_prefix + 'type=%s', FE_symbol.type.args[i])
        self.DEBUG(self.log_prefix + 'expr=%s', A)

        if not A.is_rvalue():
          A, old_A = A.to_rvalue(self, preferred = R)
          self.DEBUG(self.log_prefix + 'expr=%s', A)

          if isinstance(A.value, RegisterValue) and A.value.register.index != R.index:
            self.EMIT(MOV(R.name, A.value.name))

        elif isinstance(A.value, StringConstantValue):
          self.EMIT(LA(R.name, A.value.name))

        elif isinstance(A.value, ConstantValue):
          self.EMIT(LI(R.name, A.value.name))

        else:
          self.WARN('Unahndled func call arg load branch')

        if FE.value.storage.symbol.type.args[i] != A.type:
          A, old_A = self._perform_implicit_conversion(node.coord, A, FE.value.storage.symbol.type.args[i])
          self.DEBUG(self.log_prefix + 'expr=%s', A)

        args.append((R, arg))

        self.UP()

      self.DEBUG(self.log_prefix + 'args: %s', args)

    ret_reg = FE_symbol is not None and len(FE_symbol.type.rtype) != 0 and self.node_parent is not None and not isinstance(self.node_parent, c_ast.Compound)
    self.DEBUG(self.log_prefix + 'return type: %s', FE_symbol.type.rtype)

    if ret_reg:
      RE = Expression(value = RegisterValue(self.GET_REG(preferred = 0)), type = FE_symbol.type.rtype)

    else:
      RE = None

    self.DEBUG(self.log_prefix + 'FN retval: %s', RE)

    if isinstance(FE.value, MemorySlotValue):
      self.EMIT(CALL(FE.value.name))

    else:
      self.WARN(self.log_prefix + 'Unhandled call branch')

    if RE is not None:
      R = self.GET_REG()
      self.EMIT(MOV(R.name, RE.value.name))
      self.PUT_REG(RE.value.register)
      RE.value = RegisterValue(R)

    for R, storage in reversed(restore_regs):
      self.DEBUG(self.log_prefix + 'restore value of %s', R)
      self.EMIT(Comment('restore content of %s' % R))

      if storage is None:
        self.EMIT(POP(R.name))

      else:
        storage.unspill_register(self, R)

    self.INFO(self.log_prefix + 'parent=%s', dump_node(self.node_parent))

    self.EMIT(Comment('/ ' + comment))
    self.UP()
    return RE

  def visit_FuncDef(self, node):
    comment = dump_node(node)

    self.INFO(self.log_prefix + comment)
    self.DOWN()
    self.EMIT(Comment(comment))

    decl = node.decl

    self.INFO(self.log_prefix + 'Decl: %s', dump_node(decl))
    self.INFO(self.log_prefix + 'Type: %s', dump_node(decl.type))
    if decl.type.args is not None:
      self.INFO(self.log_prefix + 'Args: %s', dump_node(decl.type.args))
    self.INFO(self.log_prefix + 'Type type: %s', dump_node(decl.type.type))
    self.INFO(self.log_prefix + 'Return type: %s', dump_node(decl.type.type.type))

    T = CType.create_from_decl(self, decl.type)

    self.reset_scope()
    self.push_scope()

    self.FN = FN = Function(self, decl, T)
    self.functions.append(FN)

    self.GET_REG = FN.registers.get
    self.REGS    = FN.registers

    symbol = self.global_scope.get(decl.name)
    if symbol is None:
      symbol = self.global_scope.add(node.coord, Symbol(self, decl.name, FN.type))

    symbol.storage = MemorySlotStorage(symbol, decl.name)
    symbol.storage.symbol = symbol
    symbol.defined = True

    self.make_current(FN.args_block())

    if decl.type.args is not None:
      for i, arg in enumerate(decl.type.args.params):
        self.visit(arg)

        symbol = self.SCOPE.get(arg.name)
        symbol.defined = True
        self.DEBUG(self.log_prefix + 'arg symbol: %s', symbol)

        R = self.GET_REG(preferred = i + 1)
        R.dirty = True
        symbol.storage.acquire_register(R)

        # self.EMIT(STW(symbol.storage.name(), 'r%i' % (i + 1)))

        self.DEBUG(self.log_prefix + 'arg #%i: %s', i, symbol)
        self.EMIT(Comment('arg #%i: %s' % (i, symbol)))

        del FN.registers.callee_saved_regs[i + 1]

    if len(FN.type.rtype) != 0:
      del FN.registers.callee_saved_regs[0]

    self.make_current(FN.body_block())

    self.visit_Compound(node.body, create_scope = False)

    FN.finish()

    self.UP()

  def visit_ID(self, node, preferred = None, keep = None):
    comment = dump_node(node)

    self.INFO(self.log_prefix + '%s: preferred=%s, keep=%s', comment, preferred, keep)
    self.DOWN()
    self.EMIT(Comment(comment))

    symbol = self.SCOPE.get(node.name)
    self.DEBUG(self.log_prefix + 'symbol=%s', symbol)

    self.EMIT(Comment('/ ' + comment))

    RE_class = LValueExpression if symbol.const is True else MLValueExpression
    value_class = StackSlotValue if isinstance(symbol.storage, StackSlotStorage) else MemorySlotValue

    RE = RE_class(value = value_class(symbol.storage), type = PointerType(symbol.type, self))

    self.UP()
    return RE

  def visit_If(self, node):
    self.INFO(self.log_prefix + dump_node(node))
    self.DOWN()

    cond_label    = self.get_new_label(name = 'if_cond')
    iftrue_label  = self.get_new_label(name = 'if_iftrue')
    iffalse_label = self.get_new_label(name = 'if_iffalse')
    past_label    = self.get_new_label(name = 'if_past')

    self.EMIT(Comment(dump_node(node)))
    self.EMIT(Comment('  iftrue=%s, iffalse=%s, past=%s' % (iftrue_label, iffalse_label, past_label)))

    cond_block    = self.FN.block(name = cond_label)
    iftrue_block  = self.FN.block(name = iftrue_label)
    iffalse_block = self.FN.block(name = iffalse_label)

    self.BLOCK.connect(cond_block)
    cond_block.connect(iftrue_block)
    cond_block.connect(iffalse_block)

    self.make_current(cond_block)
    self.process_cond(node.cond, iftrue_label = iftrue_block.names[0], iffalse_label = iffalse_block.names[0])

    self.make_current(iftrue_block)
    self.visit(node.iftrue)

    self.make_current(iffalse_block)
    if node.iffalse is not None:
      self.visit(node.iffalse)

    past_block = self.FN.block(name = past_label)
    iftrue_block.connect(past_block)
    iffalse_block.connect(past_block)

    iftrue_block.emit(J('&' + past_label))
    iffalse_block.emit(J('&' + past_label))

    self.BLOCK.connect(past_block)
    self.EMIT(Comment('connecting %s to %s' % (self.BLOCK, past_block)))
    self.make_current(past_block)

    self.UP()
    return None

  def visit_Return(self, node):
    comment = dump_node(node)

    self.INFO(self.log_prefix + comment)
    self.DOWN()
    self.EMIT(Comment(comment))

    E = self.visit_expr(node.expr)
    self.DEBUG(self.log_prefix + 'expr=%s', E)

    if not E.is_rvalue():
      E, old_E = E.to_rvalue(self, preferred = 0)

    def __free_r0():
      R0 = self.FN.registers.all_regs[0]
      self.EMIT(Comment('free r0'))
      R0.free()
      return R0

    if isinstance(E.value, RegisterValue):
      if E.value.register.index != 0:
        R = __free_r0()
        self.EMIT(MOV(R.name, E.value.name))

      else:
        pass

    elif isinstance(E.value, StringConstantValue):
      R = __free_r0()
      self.EMIT(LA(R.name, E.value.name))

    elif isinstance(E.value, ConstantValue):
      R = __free_r0()
      self.EMIT(LI(R.name, E.value.name))

    else:
      self.WARN(self.log_prefix + 'Unhandled return branch')

    self.EMIT(J('&' + self.FN.epilog_block().names[0]))
    self.BLOCK.connect(self.FN.epilog_block())

    self.EMIT(Comment('/ ' + comment))
    self.UP()
    return None

  def visit_StructRef(self, node, **kwargs):
    comment = dump_node(node)

    self.INFO(self.log_prefix + comment)
    self.DOWN()
    self.INFO(self.log_prefix + 'name=%s, field=%s', dump_node(node.name), dump_node(node.field))

    self.EMIT(Comment(comment))

    NE = self.visit_expr(node.name)
    self.DEBUG(self.log_prefix + 'name=%s', NE)

    RE_class = NE.__class__
    self.DEBUG(self.log_prefix + 'RE class=%s', RE_class.__name__)

    if node.type == '->':
      if not isinstance(NE.type.ptr_to_type, PointerType):
        raise NotAPointerError(node.coord, NE.type)

      RE_class = MLValueExpression

      old_NE = NE
      NE = LValueExpression(value = RegisterMemorySlotValue(self.GET_REG()), type = NE.type.ptr_to_type)

      self.EMIT(LW(NE.value.name, old_NE.value.name))

    if isinstance(NE.type.ptr_to_type, PointerType):
      raise IsAPointerError(node.coord, NE.type)

    field_type = NE.type.ptr_to_type.field_type(node.field.name)
    field_offset = NE.type.ptr_to_type.field_offset(node.field.name)

    self.DEBUG(self.log_prefix + 'field_type=%s, field_offset=%s', field_type, field_offset)

    RE = None

    if isinstance(NE.value, StackSlotValue):
      RE = RE_class(value = StackSlotValue(StackSlotStorage(None, NE.value.storage.offset + field_offset)), type = PointerType(field_type, self))

    elif isinstance(NE.value, RegisterMemorySlotValue):
      RE = RE_class(value = NE.value, type = PointerType(field_type, self))
      self.EMIT(ADD(RE.value.name, field_offset))

    elif isinstance(NE.value, MemorySlotValue):
      RE = RE_class(value = RegisterValue(self.GET_REG()), type = PointerType(field_type, self))
      NE.value.storage.addrof(RE.value.register, self.EMIT)
      self.EMIT(ADD(RE.value.name, field_offset))

    else:
      self.WARN(self.log_prefix + 'Unhandled struct ref for name')

    self.EMIT(Comment('/ ' + comment))
    self.UP()
    return RE

  def visit_TypeDecl(self, node, **kwargs):
    self.INFO(self.log_prefix + dump_node(node))
    self.DOWN()

    t = CType.create_from_decl(self, node)

    self.types[node.declname] = t

    self.UP()
    return RValueExpression(type = t)

  def visit_Typedef(self, node):
    self.INFO(self.log_prefix + dump_node(node))
    self.DOWN()

    ts = self.generic_visit(node)
    T = ts[0]
    self.types[node.name] = T.type
    self.DEBUG(self.log_prefix + 'typedef new type: %s', T)

    self.UP()
    return RValueExpression(type = T)

  def visit_Typename(self, node, **kwargs):
    self.INFO(self.log_prefix + dump_node(node))
    self.DOWN()

    T = CType.get_from_decl(self, node.type)

    self.UP()
    return RValueExpression(type = T)

  def visit_UnaryOp(self, node, preferred = None, keep = None):
    comment = dump_node(node)

    self.INFO(self.log_prefix + '%s, preferred=%s, keep=%s', comment, preferred, keep)
    self.DOWN()
    self.EMIT(Comment(comment))

    keep = keep or []

    E = self.visit_expr(node.expr)
    self.DEBUG(self.log_prefix + 'expr=%s', E)

    if node.op == 'p++':
      assert E.is_mlvalue() is True

      E, old_E = E.to_rvalue(self)

      self.DEBUG(self.log_prefix + 'expr=%s', E)

      RE = RValueExpression(value = self.GET_REG(), type = E.type)

      self.EMIT(MOV(RE.value.name, E.value.name))
      self.EMIT(INC(E.value.name))

      if old_E.value.is_register_backed():
        old_E.value.backing_register().dirty = True

      else:
        if len(E.type) == 1:
          self.EMIT(STB(old_E.value.name, E.value.name))

        elif len(E.type) == 2:
          self.EMIT(STB(old_E.value.name, E.value.name))

        else:
          self.WARN(self.log_prefix + 'Unhandled save branch')

      self.DEBUG(self.log_prefix + 'postfix inc: expr=%s, rexpr=%s', E, RE)

      self.EMIT(Comment('/ ' + comment))
      self.UP()
      return RE

    if node.op == '*':
      assert isinstance(E.type, PointerType) is True

      RE = E.__class__(value = RegisterValue(self.GET_REG(preferred = preferred, keep = keep + [E.value])), type = E.type.ptr_to_type)

      if len(RE.type) == 1:
        self.EMIT(LB(RE.value.name, E.value.name))

      elif len(RE.type) == 2:
        self.EMIT(LW(RE.value.name, E.value.name))

      else:
        self.WARN(self.log_prefix + 'Unhandled deref branch')

      self.EMIT(Comment('/ ' + comment))
      self.UP()
      return RE

    if node.op == '&':
      assert E.is_lvalue() is True

      if E.value.is_register_backed():
        E.value.storage.spill_register(self)

      if isinstance(E.value, StackSlotValue):
        RE = RValueExpression(value = RegisterValue(self.GET_REG()), type = E.type)
        E.value.storage.addrof(RE.value, self.EMIT)

      elif isinstance(E.value, RegisterMemorySlotValue):
        RE = RValueExpression(value = E.value, type = E.type)

      else:
        self.WARN(self.log_prefix + 'Unhandled addrof branch')

      self.EMIT(Comment('/ ' + comment))
      self.UP()
      return RE

    if node.op == '~':
      if E.is_lvalue():
        RE, old_E = E.to_rvalue(self)

      else:
        RE = RValueExpression(value = RegisterValue(self.GET_REG()), type = E.type)

      if isinstance(E.value, StringConstantValue):
        self.EMIT(LA(RE.value.name, E.value.name))

      elif isinstance(E.value, ConstantValue):
        self.EMIT(LI(RE.value.name, E.value.name))

      self.EMIT(NOT(RE.value.name))

      self.EMIT(Comment('/ ' + comment))
      self.UP()
      return RE

    if node.op == 'sizeof':
      RE = RValueExpression(value = RegisterValue(self.GET_REG(preferred = preferred, keep = keep)), type = UnsignedIntType(self))
      self.EMIT(LI(RE.value.name, len(E.type)))

      self.UP()
      return RE

    self.WARN('Unhandled unary op: op=%s', node.op)

    self.UP()
    return None, None, None

  def visit_While(self, node):
    comment = dump_node(node)

    self.INFO(self.log_prefix + dump_node(node))
    self.DOWN()
    self.EMIT(Comment(comment))

    head_block = self.FN.block(name = self.get_new_label(name = 'while_head'))
    body_block = self.FN.block(name = self.get_new_label(name = 'while_body'))
    past_block = self.FN.block(name = self.get_new_label(name = 'while_past'))

    self.break_stack.append(past_block.names[0])
    self.continue_stack.append(head_block.names[0])

    self.BLOCK.connect(head_block)
    head_block.connect(body_block)
    head_block.connect(past_block)
    body_block.connect(head_block)
    body_block.connect(past_block)

    self.DEBUG(self.log_prefix + 'head_block: %s', head_block)

    self.make_current(head_block)

    self.process_cond(node.cond, iffalse_label = past_block.names[0])

    self.make_current(body_block)
    self.visit(node.stmt)
    self.EMIT(J('&' + head_block.names[0]))

    self.make_current(past_block)

    self.break_stack.pop()
    self.continue_stack.pop()

    self.EMIT(Comment('/ ' + comment))
    self.UP()

from . import AST_PASSES
AST_PASSES['ast-codegen'] = CodegenVisitor



import os
import sys

from six.moves import range

from ..mm import u32_t, u8_t, PAGE_SIZE
from ..mm.binary import File, SectionTypes, SectionFlags
from ..util import BinaryFile, align
from ..devices.storage import BLOCK_SIZE
from ..hdt import HDT

NULL8  = u8_t(0)
NULL32 = u32_t(0)

def align_file(logger, f_out, boundary):
  logger.debug('Adding padding bytes to align with BLOCK_SIZE')

  f_out.seek(0, os.SEEK_END)
  pos = f_out.tell()
  logger.debug('  Last position: %s => %s', pos, align(boundary, pos))

  missing = align(boundary, pos) - pos
  logger.debug('  Missing %s bytes', missing)

  for _ in range(0, missing):
    f_out.write(NULL8)

def create_binary_image(logger, f_in, f_out, bio = False):
  f_out.seek(0)

  reserved_space = 0

  if bio is True:
    logger.debug('Reserving 1st block for header')

    reserved_space = BLOCK_SIZE

    for _ in range(0, BLOCK_SIZE // 4):
      f_out.write(NULL32)

  for s_header, s_content in f_in.sections():
    s_name = f_in.string_table.get_string(s_header.name)
    s_flags = SectionFlags.from_encoding(s_header.flags)

    logger.debug('Process section %s', s_name)

    if s_header.type not in (SectionTypes.TEXT, SectionTypes.DATA):
      logger.debug('  Not a text or data section, ignore')
      continue

    if s_flags.loadable != 1:
      logger.debug('  Not loadable, ignore')
      continue

    logger.debug('  Base 0x%08X', s_header.base)
    logger.debug('  Seeking to %s', s_header.base + reserved_space)
    f_out.seek(s_header.base + reserved_space)

    if s_flags.bss == 1:
      for i in range(0, s_header.data_size):
        f_out.write(NULL8)

    else:
      for i, item in enumerate(s_content):
        f_out.write(item)

    logger.debug('%s items written', i)

  align_file(logger, f_out, BLOCK_SIZE if bio else PAGE_SIZE)

  end = f_out.tell()
  logger.debug('Last position: %s', end)

  if bio is True:
    logger.debug('Writing header')

    f_out.seek(0)
    f_out.write(u32_t(end // BLOCK_SIZE))

  f_out.flush()

def create_hdt_image(logger, file_in, f_out, options):
  from .vm import process_config_options

  config = process_config_options(logger,
                                  config_file = file_in,
                                  set_options = [(section,) + tuple(option.split('=')) for section, option in (option.split(':') for option in options.set_options)],
                                  add_options = [(section,) + tuple(option.split('=')) for section, option in (option.split(':') for option in options.add_options)],
                                  enable_devices = options.enable_devices,
                                  disable_devices = options.disable_devices)

  hdt = HDT(logger, config = config)
  hdt.create()

  f_out.seek(0)

  f_out.write(hdt.header)

  for entry in hdt.entries:
    f_out.write(entry)

  align_file(logger, f_out, PAGE_SIZE)

  f_out.flush()

def main():
  import optparse
  from . import add_common_options, parse_options

  parser = optparse.OptionParser()
  add_common_options(parser)

  group = optparse.OptionGroup(parser, 'File options')
  parser.add_option_group(group)
  group.add_option('-i', dest = 'file_in', default = None, help = 'Input file')
  group.add_option('-o', dest = 'file_out', default = None, help = 'Output file')
  group.add_option('-f', dest = 'force', default = False, action = 'store_true', help = 'Force overwrite of the output file')

  group = optparse.OptionGroup(parser, 'Image options')
  parser.add_option_group(group)
  group.add_option('-b', '--binary', dest = 'image', action = 'store_const', const = 'binary', help = 'Create image of a binary', default = 'binary')
  group.add_option('-H', '--hdt',    dest = 'image', action = 'store_const', const = 'hdt',    help = 'Create image of HDT')
  group.add_option('--bio',          dest = 'bio',   action = 'store_true',  default = False,  help = 'Allow loading by BIO')

  group = optparse.OptionGroup(parser, 'Machine config options')
  parser.add_option_group(group)
  group.add_option('--set-option', dest = 'set_options', action = 'append', default = [], metavar = 'SECTION:OPTION=VALUE', help = 'Set option')
  group.add_option('--add-option', dest = 'add_options', action = 'append', default = [], metavar = 'SECTION:OPTION=VALUE', help = 'Add value to an option')
  group.add_option('--enable-device', dest = 'enable_devices', action = 'append', default = [], metavar = 'DEVICE', help = 'Enable device')
  group.add_option('--disable-device', dest = 'disable_devices', action = 'append', default = [], metavar = 'DEVICE', help = 'Disable device')

  options, logger = parse_options(parser)

  if not options.file_in or not options.file_out:
    parser.print_help()
    sys.exit(1)

  if options.image == 'binary':
    with File.open(logger, options.file_in, 'r') as f_in:
      f_in.load()

      with BinaryFile.open(logger, options.file_out, 'w') as f_out:
        create_binary_image(logger, f_in, f_out, bio = options.bio)

  elif options.image == 'hdt':
    with BinaryFile.open(logger, options.file_out, 'w') as f_out:
      create_hdt_image(logger, options.file_in, f_out, options)



"""
VM configuration management
"""

import functools

from six import iteritems
from six.moves import StringIO
from six.moves.configparser import ConfigParser, NoSectionError, NoOptionError

from .util import str2int

def bool2option(b):
  """
  Get config-file-usable string representation of boolean value.

  :param bool b: value to convert.
  :rtype: string
  :returns: ``yes`` if input is ``True``, ``no`` otherwise.
  """

  return 'yes' if b else 'no'

class MachineConfig(ConfigParser):
  """
  Contains configuration of the whole VM, and provides methods for parsing,
  inspection and extending this configuration.
  """

  def __init__(self, *args, **kwargs):
    ConfigParser.__init__(self, *args, **kwargs)

    self.breakpoints_cnt = 0
    self.mmaps_cnt       = 0
    self.devices_cnt     = 0

  def get(self, section, option, default = None, **kwargs):
    """
    Get value for an option.

    :param string section: config section.
    :param string option: option name,
    :param default: this value will be returned, if no such option exists.
    :rtype: string
    :returns: value of config option.
    """

    try:
      return ConfigParser.get(self, section, option, **kwargs)

    except (NoSectionError, NoOptionError):
      return default

  def set(self, section, option, value, *args, **kwargs):
    ConfigParser.set(self, section, option, str(value), *args, **kwargs)

  def getint(self, section, option, default = None):
    try:
      v = ConfigParser.get(self, section, option)
      return str2int(v) if v is not None else default

    except (NoSectionError, NoOptionError):
      return default

  _boolean_states = {'1': True, 'yes': True, 'true': True, 'on': True, '0': False, 'no': False, 'false': False, 'off': False}

  def getbool(self, section, option, default = None):
    try:
      v = ConfigParser.get(self, section, option).lower()

      if v is None:
        return default

      if v not in self._boolean_states:
        raise ValueError('Not a boolean: {}'.format(v))

      return self._boolean_states[v]

    except (NoSectionError, NoOptionError):
      return default

  def getfloat(self, section, option, default = None):
    try:
      v = ConfigParser.get(self, section, option)
      return float(v) if v is not None else default

    except (NoSectionError, NoOptionError):
      return default

  def create_getters(self, section):
    return (functools.partial(self.get, section), functools.partial(self.getbool, section), functools.partial(self.getint, section))

  def read(self, *args, **kwargs):
    ConfigParser.read(self, *args, **kwargs)

    self.__count_breakpoints()
    self.__count_mmaps()
    self.__count_devices()

  def dumps(self):
    s = StringIO()
    self.write(s)
    return s.getvalue()

  def __sections_with_prefix(self, prefix):
    return [s_name for s_name in self.sections() if s_name.startswith(prefix)]

  def __count(self, prefix):
    return len(self.__sections_with_prefix(prefix))

  def __count_breakpoints(self):
    self.breakpoints_cnt = self.__count('breakpoint-')

  def __count_mmaps(self):
    self.mmaps_cnt = self.__count('mmap-')

  def __count_devices(self):
    self.devices_cnt = self.__count('device-')

  def iter_breakpoints(self):
    for s_name in self.__sections_with_prefix('breakpoint-'):
      yield s_name

  def iter_mmaps(self):
    for s_name in self.__sections_with_prefix('mmap-'):
      yield s_name

  def iter_devices(self):
    for s_name in self.__sections_with_prefix('device-'):
      yield s_name

  def iter_storages(self):
    for s_name in self.__sections_with_prefix('device-'):
      if self.get(s_name, 'klass') != 'storage':
        continue

      yield s_name

  def add_breakpoint(self, core, address, active = None, flip = None, ephemeral = None, countdown = None):
    bp_section = 'breakpoint-{}'.format(self.breakpoints_cnt)
    self.breakpoints_cnt += 1

    self.add_section(bp_section)
    self.set(bp_section, 'core', core)
    self.set(bp_section, 'address', address)

    if active is not None:
      self.set(bp_section, 'active', bool2option(active))

    if flip is not None:
      self.set(bp_section, 'flip', bool2option(flip))

    if ephemeral is not None:
      self.set(bp_section, 'ephemeral', bool2option(ephemeral))

    if countdown is not None:
      self.set(bp_section, 'countdown', str(countdown))

  def add_mmap(self, filepath, address, size, offset = None, access = None, shared = None):
    mmap_section = 'mmap-{}'.format(self.mmaps_cnt)
    self.mmaps_cnt += 1

    self.add_section(mmap_section)
    self.set(mmap_section, 'file', filepath)
    self.set(mmap_section, 'address', address)
    self.set(mmap_section, 'size', size)

    if offset is not None:
      self.set(mmap_section, 'offset', offset)

    if access is not None:
      self.set(mmap_section, 'access', access)

    if shared is not None:
      self.set(mmap_section, 'shared', shared)

  def add_device(self, klass, driver, **kwargs):
    """
    Add another device to the configuration.

    :param string klass: class of the device (``klass`` option).
    :param string driver: device driver - dot-separated path to class (``driver``
      option).
    :param kwargs: all keyword arguments will be added to the section as device
      options.
    """

    section = 'device-{}'.format(self.devices_cnt)
    self.devices_cnt += 1

    self.add_section(section)
    self.set(section, 'klass', klass)
    self.set(section, 'driver', driver)

    for name, value in iteritems(kwargs):
      self.set(section, name, value)

    return section

  def add_storage(self, driver, sid, filepath = None):
    """
    Add another storage to the configuration.

    :param string driver: storage's driver - dot-separated path to class (``driver`` option).
    :param int sid: storage's SID (``sid`` options).
    :param string filepath: path to backend file, if there's any (``filepath`` option).
    """

    self.add_device('storage', driver, sid = sid, filepath = filepath)



import enum

from ...interfaces import ISnapshotable
from ..import CPUException
from . import Coprocessor
from ...errors import AccessViolationError
from ...mm import u32_t
from ...util import Flags

class ReadOnlyRegisterError(CPUException):
  def __init__(self, r, *args, **kwargs):
    super(ReadOnlyRegisterError, self).__init__('Register cr{:d} is read-only.'.format(r), *args, **kwargs)

class WriteOnlyRegisterError(CPUException):
  def __init__(self, r, *args, **kwargs):
    super(WriteOnlyRegisterError, self).__init__('Register cr{:d} is write-only.'.format(r), *args, **kwargs)

class ControlRegisters(enum.IntEnum):
  CR0 = 0  # CPUID
  CR1 = 1  # Interrupt Vector Table address
  CR2 = 2  # Page Table address
  CR3 = 3  # Flags

class CoreFlags(Flags):
  _flags = ['pt_enabled', 'jit']
  _labels = 'PJ'

class ControlCoprocessor(ISnapshotable, Coprocessor):
  def read_cr0(self):
    return u32_t((self.core.cpu.id << 16) | self.core.id).value

  def read_cr1(self):
    return u32_t(self.core.ivt_address).value

  def write_cr1(self, address):
    self.core.ivt_address = address

  def read_cr2(self):
    return u32_t(self.core.mmu.pt_address).value

  def write_cr2(self, address):
    self.core.mmu.pt_address = address

  def read_cr3(self):
    return CoreFlags.create(pt_enabled = self.core.mmu.pt_enabled, jit = self.core.cpu.machine.config.getbool('machine', 'jit', False)).to_int()

  def write_cr3(self, value):
    flags = CoreFlags.from_int(value)

    self.core.mmu.pt_enabled = flags.pt_enabled

  def read(self, r):
    if not self.core.privileged:
      raise AccessViolationError('It is not allowed to read control registers in non-privileged mode')

    handler = 'read_cr%i' % (r.value if isinstance(r, ControlRegisters) else r)

    if not hasattr(self, handler):
      raise WriteOnlyRegisterError(r)

    return getattr(self, handler)()

  def write(self, r, value):
    if not self.core.privileged:
      raise AccessViolationError('It is not allowed to modify control registers in non-privileged mode')

    handler = 'write_cr%i' % (r.value if isinstance(r, ControlRegisters) else r)

    if not hasattr(self, handler):
      raise ReadOnlyRegisterError(r)

    return getattr(self, handler)(value)



from setuptools import setup, Extension
from setuptools.command.test import test as TestCommand
import sys

mod_native_data_cache = Extension('ducky.native.data_cache', sources = ['ducky/native/data_cache.c'])

install_requires = [
  'colorama',
  'enum34',
  'tabulate',
  'pycparser',
  'six',
  'mako',
  'Twisted',
  'autobahn'
]

tests_requires = [
  'tox',
  'virtualenv'
]

class Tox(TestCommand):
  user_options = [('tox-args=', 'a', 'Arguments to pass to tox')]

  def initialize_options(self):
    TestCommand.initialize_options(self)
    self.tox_args = None

  def finalize_options(self):
    TestCommand.finalize_options(self)
    self.test_args = []
    self.test_suite = True

  def run_tests(self):
    import tox
    import shlex

    args = self.tox_args
    if args:
      args = shlex.split(self.tox_args)

    errno = tox.cmdline(args=args)
    sys.exit(errno)

setup(name = 'ducky',
      version = '3.0',
      description = 'Simple virtual CPU/machine simulator',
      long_description = 'Ducky is a simple virtual CPU/machine simulator, with modular design and interesting features.',
      url = 'https://github.com/happz/ducky',
      download_url = 'https://github.com/happz/ducky/tarball/3.0',
      author = 'Milos Prchlik',
      author_email = 'happz@happz.cz',
      license = 'MIT',
      classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Forth',
        'Programming Language :: Assembly',
        'Topic :: Software Development :: Assemblers',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Compilers',
        'Topic :: Software Development :: Interpreters'
      ],
      keywords = 'virtual CPU simulator',
      packages = [
        'ducky',
        'ducky.cc',
        'ducky.cc.passes',
        'ducky.cpu',
        'ducky.cpu.coprocessor',
        'ducky.mm',
        'ducky.devices',
        'ducky.tools'
      ],
      entry_points = {
        'console_scripts': [
          'ducky-as = ducky.tools.as:main',
          'ducky-cc = ducky.tools.cc:main',
          'ducky-ld = ducky.tools.ld:main',
          'ducky-vm = ducky.tools.vm:main',
          'ducky-objdump = ducky.tools.objdump:main',
          'ducky-coredump = ducky.tools.coredump:main',
          'ducky-profile = ducky.tools.profile:main',
          'ducky-img = ducky.tools.img:main',
          'ducky-defs = ducky.tools.defs:main'
        ]
      },
      package_dir = {'ducky': 'ducky'},
      zip_safe = False,
      install_requires = install_requires,
      tests_require = tests_requires,
      cmdclass = {
        'test': Tox
      }
     )



"""
Keyboard controller - provides events for pressed and released keys.
"""

import enum
import io

from . import IRQProvider, IOProvider, DeviceFrontend, DeviceBackend, IRQList
from ..errors import InvalidResourceError
from ..mm import UINT16_FMT

DEFAULT_IRQ = 0x01
DEFAULT_PORT_RANGE = 0x100


class ControlMessages(enum.IntEnum):
  HALT = 1025

  CONTROL_MESSAGE_FIRST = 1024

class Frontend(DeviceFrontend):
  def __init__(self, machine, name):
    super(Frontend, self).__init__(machine, 'input', name)

    self._comm_queue = machine.comm_channel.get_queue(name)
    self._streams = []
    self._stream = None

    self.backend = machine.get_device_by_name(name)

  @staticmethod
  def create_from_config(machine, config, section):
    slave = config.get(section, 'slave', default = section)

    return Frontend(machine, slave)

  def boot(self):
    super(Frontend, self).boot()

    self._open_input()
    self.backend.boot()

  def halt(self):
    self._close_input()
    self.backend.halt()

    super(Frontend, self).halt()

  def enqueue_stream(self, stream):
    self.machine.DEBUG('%s.enqueue_input: stream=%s', self.__class__.__name__, stream)

    if not stream.has_poll_support():
      raise InvalidResourceError('Keyboard stream must support polling')

    self._streams.append(stream)

  def _close_input(self):
    self.machine.DEBUG('%s._close_input: input=%s', self.__class__.__name__, self._stream)

    if self._stream is None:
      return

    self._stream.unregister_with_reactor(self.machine.reactor)
    self._stream = None

  def _open_input(self):
    self.machine.DEBUG('%s._open_input', self.__class__.__name__)

    self._close_input()

    if not self._streams:
      self.machine.DEBUG('%s._open_input: no additional input streams', self.__class__.__name__)
      self._comm_queue.write_in(ControlMessages.HALT)

      # if not self.queue or self.queue[-1] != ControlMessages.HALT:
      #   self.machine.DEBUG('signal halt')
      #   self.queue.append(ControlMessages.HALT)
      return

    self._stream = self._streams.pop(0)
    self.machine.DEBUG('%s._open_input: stream=%r', self.__class__.__name__, self._stream)

    self._stream.register_with_reactor(self.machine.reactor, on_read = self._handle_raw_input, on_error = self._handle_input_error)

  def _handle_input_error(self):
    self.machine.DEBUG('%s._handle_input_error')

    self._open_input()

  def _handle_raw_input(self):
    self.machine.DEBUG('%s._handle_raw_input', self.__class__.__name__)

    assert self._stream is not False

    buff = self._stream.read(size = io.DEFAULT_BUFFER_SIZE)
    self.machine.DEBUG('%s._handle_raw_input: buff=%s (%s)', self.__class__.__name__, buff, type(buff))

    if buff is None:
      self.machine.DEBUG('%s._handle_raw_input: nothing to do, no input', self.__class__.__name__)
      return

    if not buff:
      # EOF
      self._open_input()
      return

    self.machine.DEBUG('%s._handle_raw_input: adding %i chars', self.__class__.__name__, len(buff))

    self._comm_queue.write_in(buff)

    self.machine.trigger_irq(self.backend)

class Backend(IRQProvider, IOProvider, DeviceBackend):
  def __init__(self, machine, name, port = None, irq = None):
    super(Backend, self).__init__(machine, 'input', name)

    self.port = port or DEFAULT_PORT_RANGE
    self.ports = [port]
    self.irq = irq or DEFAULT_IRQ

    self._comm_queue = machine.comm_channel.create_queue(name)
    self._key_queue = []

  @staticmethod
  def create_from_config(machine, config, section):
    return Backend(machine, section,
                   port = config.getint(section, 'port', DEFAULT_PORT_RANGE),
                   irq = config.getint(section, 'irq', IRQList.KEYBOARD))

  def __repr__(self):
    return 'basic keyboard controller on [%s] as %s' % (', '.join([UINT16_FMT(port) for port in self.ports]), self.name)

  def boot(self):
    self.machine.DEBUG('%s.boot', self.__class__.__name__)

    for port in self.ports:
      self.machine.register_port(port, self)

    self.machine.tenh('hid: %s', self)

  def halt(self):
    self.machine.DEBUG('%s.halt', self.__class__.__name__)

    for port in self.ports:
      self.machine.unregister_port(port)

  def _process_input_event(self, e):
    self.machine.DEBUG('%s.__process_input_event: e=%r', self.__class__.__name__, e)

    if isinstance(e, (list, bytearray, bytes)):
      for key in e:
        self._key_queue.append(key)

    elif isinstance(e, ControlMessages):
      self._key_queue.append(e)

    else:
      raise InvalidResourceError('Unknown message: e=%s, type=%s' % (e, type(e)))

  def _process_input_events(self):
    self.machine.DEBUG('%s.__process_input_events', self.__class__.__name__)

    while True:
      e = self._comm_queue.read_in()
      if e is None:
        return

      self._process_input_event(e)

  def _read_char(self):
    self.machine.DEBUG('%s._read_char', self.__class__.__name__)

    def __do_read_char():
      try:
        b = self._key_queue.pop(0)

      except IndexError:
        self.machine.DEBUG('%s._read_char: no available chars in queue', self.__class__.__name__)
        return None

      return b

    def __process_char(b):
      self.machine.DEBUG('%s._read_char: queue now has %i bytes', self.__class__.__name__, len(self._key_queue))

      if b == ControlMessages.HALT:
        self.machine.DEBUG('%s._read_char: planned halt, execute', self.__class__.__name__)
        self.machine.halt()

        return None

      self.machine.DEBUG('%s._read_char: c=%s ()', self.__class__.__name__, b)

      return b

    b = __do_read_char()
    if b is not None:
      return __process_char(b)

    self._process_input_events()

    b = __do_read_char()
    if b is None:
      return None

    return __process_char(b)

  def read_u8(self, port):
    self.machine.DEBUG('%s.read_u8: port=%s', self.__class__.__name__, UINT16_FMT(port))

    if port not in self.ports:
      raise InvalidResourceError('Unhandled port: %s' % UINT16_FMT(port))

    b = self._read_char()
    if not b:
      self.machine.DEBUG('%s.read_u8: empty input, signal it downstream', self.__class__.__name__)
      return 0xFF

    self.machine.DEBUG('%s.read_u8: input byte is %i', self.__class__.__name__, b)

    return b



from six import itervalues

from . import BlockVisitor
from .. import J

class BlockTreeSimplifyVisitor(BlockVisitor):
  priority = 500

  def do_visit_fn(self, fn):
    # Remove empty blocks
    for block in fn.blocks[:]:
      # self.DEBUG(self.log_prefix + 'Block: %s', block)
      # self.DEBUG(self.log_prefix + '  Code: %s, Instr: %s', block.code, block.instructions())

      if block.instructions():
        continue

      if block.names:
        # If it has no outgoing block, it weird but let not mess with it right now
        # if it has multiple outgoing blocks, lets not mess with this one
        if len(block.outgoing) != 1:
          continue

        next_block = list(block.outgoing.values())[0]

        self.DEBUG(self.log_prefix + 'Block %s is empty, but is named - merging its names to the following block %s', block, next_block)

        # Merge names to the following block
        next_block.names += block.names
        block.names = []

      if block.names or len(block.outgoing) > 1:
        continue

      fn.blocks.remove(block)

      out_block = list(block.outgoing.values())[0] if block.outgoing else None
      if out_block is not None:
        del out_block.incoming[block.id]

      for in_block in itervalues(block.incoming):
        del in_block.outgoing[block.id]
        if out_block is not None:
          in_block.add_outgoing(out_block)
          out_block.add_incoming(in_block)

      self.DEBUG(self.log_prefix + 'Block %s is empty, removed', block)

    # Get rid of "jump-to-directly-next-block" instructions
    prev = None
    for block in fn.blocks:
      if prev is None or not prev.code or not isinstance(prev.instructions()[-1], J) or prev.instructions()[-1].operands[0] not in ['&%s' % name for name in block.names]:
        prev = block
        continue

      prev.code.remove(prev.instructions()[-1])
      self.DEBUG(self.log_prefix + 'Block %s jumps directly to next block %s', prev, block)

      prev = block

from . import BT_PASSES
BT_PASSES['bt-simplify'] = BlockTreeSimplifyVisitor



"""
Persistent storage support.

Several different persistent storages can be attached to a virtual machine, each
with its own id. This module provides methods for manipulating their content. By
default, storages operate with blocks of constant, standard size, though this is
not a mandatory requirement - storage with different block size, or even with
variable block size can be implemented.

Each block has its own id. Block IO operations read or write one or more blocks to
or from a device. IO is requested by invoking the virtual interrupt, with properly
set values in registers.
"""

import os
import six

from ..errors import InvalidResourceError
from . import Device, IRQList, IRQProvider, IOProvider
from ..mm import u32_t
from ..util import UINT16_FMT, UINT32_FMT, F

#: Size of block, in bytes.
BLOCK_SIZE = 1024

DEFAULT_PORT_RANGE = 0x400
DEFAULT_IRQ = IRQList.BIO

PORT_RANGE = 0x0005

BIO_RDY   = 0x00000001
BIO_ERR   = 0x00000002
BIO_READ  = 0x00000004
BIO_WRITE = 0x00000008
BIO_BUSY  = 0x00000010
BIO_DMA   = 0x00000020
BIO_SRST  = 0x00000040

BIO_USER  = BIO_READ | BIO_WRITE | BIO_DMA | BIO_SRST

class StorageAccessError(Exception):
  """
  Base class for storage-related exceptions.
  """

  pass

class Storage(Device):
  """
  Base class for all block storages.

  :param ducky.machine.Machine machine: machine storage is attached to.
  :param int sid: id of storage.
  :param int size: size of storage, in bytes.
  """

  def __init__(self, machine, name, sid = None, size = None, *args, **kwargs):
    super(Storage, self).__init__(machine, 'storage', name, *args, **kwargs)

    self.sid = sid
    self.size = size

  def do_read_blocks(self, start, cnt):
    """
    Read one or more blocks from device to internal buffer.

    Child classes are supposed to reimplement this particular method.

    :param u32_t start: index of the first requested block.
    :param u32_t cnt: number of blocks to read.
    """

    raise NotImplementedError()

  def do_write_blocks(self, start, cnt, buff):
    """
    Write one or more blocks from internal buffer to device.

    Child classes are supposed to reimplement this particular method.

    :param u32_t start: index of the first requested block.
    :param u32_t cnt: number of blocks to write.
    """

    raise NotImplementedError()

  def read_blocks(self, start, cnt):
    """
    Read one or more blocks from device to internal buffer.

    Child classes should not reimplement this method, as it provides checks
    common for (probably) all child classes.

    :param u32_t start: index of the first requested block.
    :param u32_t cnt: number of blocks to read.
    """

    self.machine.DEBUG('%s.read_blocks: id=%s, start=%s, cnt=%s', self.__class__.__name__, self.sid, start, cnt)

    if (start + cnt) * BLOCK_SIZE > self.size:
      raise StorageAccessError('Out of bounds access: storage size {} is too small'.format(self.size))

    return self.do_read_blocks(start, cnt)

  def write_blocks(self, start, cnt, buff):
    """
    Write one or more blocks from internal buffer to device.

    Child classes should not reimplement this method, as it provides checks
    common for (probably) all child classes.

    :param u32_t start: index of the first requested block.
    :param u32_t cnt: number of blocks to write.
    """

    self.machine.DEBUG('%s.write_blocks: id=%s, start=%s, cnt=%s', self.__class__.__name__, self.sid, start, cnt)

    if (start + cnt) * BLOCK_SIZE > self.size:
      raise StorageAccessError('Out of bounds access: storage size {} is too small'.format(self.size))

    self.do_write_blocks(start, cnt, buff)

class FileBackedStorage(Storage):
  """
  Storage that saves its content into a regular file.
  """

  def __init__(self, machine, name, filepath = None, *args, **kwargs):
    """
    :param machine.Machine machine: virtual machine this storage is attached to
    :param int sid: storage id
    :param path: path to a underlying file
    """

    self.filepath = filepath
    st = os.stat(filepath)

    super(FileBackedStorage, self).__init__(machine, name, size = st.st_size, *args, **kwargs)

    self.filepath = filepath
    self.file = None

  @staticmethod
  def create_from_config(machine, config, section):
    return FileBackedStorage(machine, section, sid = config.getint(section, 'sid', None), filepath = config.get(section, 'filepath', None))

  def boot(self):
    self.machine.DEBUG('FileBackedStorage.boot')

    self.file = open(self.filepath, 'r+b')

    self.machine.tenh('storage: file %s as storage #%i (%s)', self.filepath, self.sid, self.name)

  def halt(self):
    self.machine.DEBUG('FileBackedStorage.halt')

    self.file.flush()
    self.file.close()

  if six.PY2:
    def _read(self, cnt):
      return bytearray([ord(c) for c in self.file.read(cnt)])

    def _write(self, buff):
      self.file.write(''.join([chr(b) for b in buff]))

  else:
    def _read(self, cnt):
      return self.file.read(cnt)

    def _write(self, buff):
      self.file.write(bytes(buff))

  def do_read_blocks(self, start, cnt):
    self.machine.DEBUG('%s.do_read_blocks: start=%s, cnt=%s', self.__class__.__name__, start, cnt)

    self.file.seek(start * BLOCK_SIZE)

    return self._read(cnt * BLOCK_SIZE)

  def do_write_blocks(self, start, cnt, buff):
    self.machine.DEBUG('%s.do_write_blocks: start=%s, cnt=%s', self.__class__.__name__, start, cnt)

    self.file.seek(start * BLOCK_SIZE)
    self._write(buff)
    self.file.flush()

class BlockIO(IRQProvider, IOProvider, Device):
  def __init__(self, machine, name, port = None, irq = None, *args, **kwargs):
    super(BlockIO, self).__init__(machine, 'bio', name, *args, **kwargs)

    self._device    = u32_t(0xFFFFFFFF)
    self._flags     = u32_t(BIO_RDY)
    self._block     = u32_t(0x00000000)
    self._count     = u32_t(0x00000000)
    self._address   = u32_t(0x00000000)

    self.port = port
    self.ports = list(range(port, port + PORT_RANGE))

    self.reset()

  @staticmethod
  def create_from_config(machine, config, section):
    return BlockIO(machine,
                   section,
                   port = config.getint(section, 'port', DEFAULT_PORT_RANGE),
                   irq = config.getint(section, 'irq', IRQList.TIMER))

  def boot(self):
    self.machine.DEBUG('BIO.boot')

    for port in self.ports:
      self.machine.register_port(port, self)

    self.machine.tenh('BIO: controller on [%s] as %s', UINT16_FMT(self.port), self.name)

  def halt(self):
    for port in self.ports:
      self.machine.unregister_port(port)

  def reset(self):
    self.storage = None
    self.start = None
    self.count = None
    self.address = None

    self.buffer = None
    self.buffer_length = None
    self.buffer_index = None

    self.dma = False
    self.busy = False

    self._flags = u32_t(BIO_RDY)

  def buff_to_memory(self, addr, buff):
    self.machine.DEBUG('%s.buff_to_memory: addr=%s', self.__class__.__name__, UINT32_FMT(addr))

    for i in range(0, len(buff)):
      self.machine.memory.write_u8(addr + i, buff[i])

  def memory_to_buff(self, addr, length):
    self.machine.DEBUG('%s.memory_to_buff: addr=%s, length=%s', self.__class__.__name__, UINT32_FMT(addr), length)

    return bytearray([self.machine.memory.read_u8(addr + i) for i in range(0, length)])

  def __flag_busy(self):
    self._flags.value |=  BIO_BUSY
    self._flags.value &= ~BIO_RDY

  def __flag_finished(self):
    self._flags.value &= ~BIO_BUSY
    self._flags.value |= BIO_RDY

  def __flag_error(self):
    self._flags.value &= ~BIO_RDY
    self._flags.value &= ~BIO_BUSY
    self._flags.value |= BIO_ERR

  def read_u32(self, port):
    self.machine.DEBUG('%s.read_u32: port=%s', self.__class__.__name__, UINT16_FMT(port))

    if port not in self.ports:
      raise InvalidResourceError(F('Unhandled port: {port:S}', port = port))

    port -= self.port

    if port == 0x0000:
      return self._flags.value

    if port == 0x0005:
      if self.buffer_index == self.buffer_length:
        self._flags |= BIO_RDY
        return 0xFFFFFFFF

      v =  (self.buffer[self.buffer_index + 3] << 24) | (self.buffer[self.buffer_index + 2] << 16) | (self.buffer[self.buffer_index + 1] <<  8) |  self.buffer[self.buffer_index]

      self.buffer_index += 4
      return v

    raise InvalidResourceError(F('Write-only port: {port:S}', port = port + self.port))

  def write_u32(self, port, value):
    self.machine.DEBUG('%s.write_u32: port=%s, value=%s', self.__class__.__name__, UINT16_FMT(port), UINT32_FMT(value))

    if port not in self.ports:
      raise InvalidResourceError(F('Unhandled port: {port:S}', port = port))

    port -= self.port
    value &= 0xFFFFFFFF

    if port == 0x0000:
      value &= BIO_USER

      if value & BIO_SRST:
        self.machine.DEBUG('%s.write_u32: SRST', self.__class__.__name__)
        self.reset()

      if value & BIO_DMA:
        self.machine.DEBUG('%s.write_u32: DMA', self.__class__.__name__)
        self.dma = True

      if value & BIO_READ or value & BIO_WRITE:
        self.machine.DEBUG('%s.write_u32: IO', self.__class__.__name__)

        self.__flag_busy()

        self.buffer_index = 0

        if value & BIO_READ:
          try:
            self.buffer = self.storage.read_blocks(self._block.value, self._count.value)

          except StorageAccessError:
            self.__flag_error()

          else:
            if self.dma:
              self.buff_to_memory(self._address.value, self.buffer)
              self.__flag_finished()

        else:
          if self.dma is True:
            self.buffer = self.memory_to_buff(self._address.value, self._count.value * BLOCK_SIZE)

            try:
              self.storage.write_blocks(self._block.value, self._count.value, self.buffer)

            except StorageAccessError:
              self.__flag_error()

            else:
              self.__flag_finished()

      return

    if port == 0x0001:
      try:
        self.storage = self.machine.get_storage_by_id(value)
        self._device_id = value

      except InvalidResourceError:
        self.__flag_error()

      return

    if port == 0x0002:
      self._block.value = value
      return

    if port == 0x0003:
      self._count.value = value
      return

    if port == 0x0004:
      self._address.value = value
      return

    if port == 0x0005:
      if self.buffer_index == self.buffer_length:
        self._flags.value |= BIO_ERR
        return

      self.buffer[self.buffer_index]     =  value        & 0xFF
      self.buffer[self.buffer_index + 1] = (value >>  8) & 0xFF
      self.buffer[self.buffer_index + 1] = (value >> 16) & 0xFF
      self.buffer[self.buffer_index + 1] = (value >> 24) & 0xFF
      self.buffer_index += 4
      return

    raise InvalidResourceError(F('Read-only port: {port:S}', port = port + self.port))



"""
:py:class:`ducky.machine.Machine` is *the* virtual machine. Each instance
represents self-contained virtual machine, with all its devices, memory, CPUs
and other necessary properties.
"""

import itertools
import collections
import os
import sys
import time

from six import itervalues, iteritems
from collections import defaultdict, OrderedDict

from . import mm
from . import snapshot

from . import __version__

from .interfaces import IMachineWorker, ISnapshotable, IReactorTask

from .console import ConsoleMaster
from .errors import InvalidResourceError
from .log import create_logger
from .mm import UINT16_FMT
from .reactor import Reactor
from .snapshot import SnapshotNode
from .util import F
from .boot import ROMLoader

from functools import partial

class MachineState(SnapshotNode):
  def __init__(self):
    super(MachineState, self).__init__('nr_cpus', 'nr_cores')

  def get_cpu_states(self):
    return [__state for __name, __state in iteritems(self.get_children()) if __name.startswith('cpu')]

  def get_cpu_state_by_id(self, cpuid):
    return self.get_children()['cpu{}'.format(cpuid)]

class CommQueue(object):
  def __init__(self, channel):
    self.channel = channel

    self.queue_in = []
    self.queue_out = []

  def is_empty_out(self):
    return not bool(self.queue_out)

  def is_empty_in(self):
    return not bool(self.queue_in)

  def write_out(self, o):
    self.queue_out.append(o)

  def write_in(self, o):
    self.queue_in.append(o)

  def __read(self, queue):
    try:
      return queue.pop(0)

    except IndexError:
      return None

  def read_out(self):
    return self.__read(self.queue_out)

  def read_in(self):
    return self.__read(self.queue_in)

class CommChannel(object):
  def __init__(self, machine):
    self.machine = machine

    self._queues = {}

  def create_queue(self, name):
    queue = CommQueue(self)
    self._queues[name] = queue
    return queue

  def get_queue(self, name):
    return self._queues[name]

  def unregister_queue(self, name):
    del self._queues[name]


class IRQRouterTask(IReactorTask):
  """
  This task is responsible for distributing triggered IRQs between CPU cores.
  When IRQ is triggered, IRQ source (i.e. device that requires attention) is
  appended to this tasks queue (:py:attr:`ducky.machine.IRQRouterTask.qeueu`).
  As long as this queue is not empty, this task pops IRQ sources, selects
  free CPU core, and by calling its :py:meth:`ducky.cpu.CPUCore.irq` method
  core takes reponsibility for executing interrupt routine.

  :param ducky.machine.Machine machine: machine this task belongs to.
  """

  def __init__(self, machine):
    self.machine = machine

    from .devices import IRQList
    self.queue = [False for _ in range(0, IRQList.IRQ_COUNT)]

  def run(self):
    self.machine.DEBUG('irq: router has %i waiting irqs', self.queue.count(True))

    for irq, triggered in enumerate(self.queue):
      if triggered is not True:
        continue

      self.machine.DEBUG('irq: triggered %i', irq)
      for core in self.machine.living_cores:
        if core.hwint_allowed is not True:
          self.machine.DEBUG('irq: %s hwint not allowed', core.cpuid)
          continue

        self.machine.DEBUG('irq: interrupt %s', core.cpuid)

        self.queue[irq] = False
        core.irq(irq)
        break

      else:
        break

    if not any(self.queue):
      self.machine.reactor.task_suspended(self)

class HaltMachineTask(IReactorTask):
  def __init__(self, machine):
    self.machine = machine

  def run(self):
    self.machine.halt()

class EventBus(object):
  def __init__(self, machine):
    super(EventBus, self).__init__()

    self.machine = machine

    self.listeners = defaultdict(OrderedDict)

  def add_listener(self, event, callback, *args, **kwargs):
    self.machine.DEBUG('%s.add_listener: event=%s, callback=%s, args=%s, kwargs=%s', self.__class__.__name__, event, callback, args, kwargs)

    self.listeners[event][callback] = (args, kwargs)

  def remove_listener(self, event, callback):
    self.machine.DEBUG('%s.remove_listener: event=%s, callback=%s', self.__class__.__name__, event, callback)

    del self.listeners[event][callback]

  def trigger(self, event, *args, **kwargs):
    self.machine.DEBUG('%s.trigger: event=%s, args=%s, kwargs=%s', self.__class__.__name__, event, args, kwargs)

    for listener, (_args, _kwargs) in iteritems(self.listeners[event]):
      _args = _args + args
      _kwargs = _kwargs.copy()
      _kwargs.update(kwargs)
      listener(*args, **kwargs)

class Machine(ISnapshotable, IMachineWorker):
  """
  Virtual machine itself.
  """

  def core(self, cid):
    """
    Find CPU core by its string id.

    :param string cid: id of searched CPU core, in the form `#<cpuid>:#<coreid>`.
    :rtype: :py:class:`ducky.cpu.CPUCore`
    :returns: found core
    :raises ducky.errors.InvalidResourceError: when no such core exists.
    """

    for _cpu in self.cpus:
      for _core in _cpu.cores:
        if '#%i:#%i' % (_cpu.id, _core.id) == cid:
          return _core

    raise InvalidResourceError(F('No such CPU core: cid={cid}', cid = cid))

  def __init__(self, logger = None, stdin = None, stdout = None, stderr = None):
    self.stdin  = stdin or sys.stdin
    self.stdout = stdout or sys.stdout
    self.stderr = stderr or sys.stderr

    self.reactor = Reactor(self)

    # Setup logging
    self.LOGGER = logger or create_logger()
    self.DEBUG = self.LOGGER.debug
    self.INFO = self.LOGGER.info
    self.WARN = self.LOGGER.warning
    self.ERROR = self.LOGGER.error
    self.EXCEPTION = self.LOGGER.exception

    self._tenh = None
    self._tenh_device = None
    self._tenh_enabled = False

    self.console = ConsoleMaster(self)
    self.console.register_command('halt', cmd_halt)
    self.console.register_command('boot', cmd_boot)
    self.console.register_command('run', cmd_run)
    self.console.register_command('snap', cmd_snapshot)

    self.irq_router_task = IRQRouterTask(self)
    self.reactor.add_task(self.irq_router_task)

    self.check_living_cores_task = HaltMachineTask(self)
    self.reactor.add_task(self.check_living_cores_task)

    self.comm_channel = CommChannel(self)

    self.events = EventBus(self)

    self.living_cores = []

    self.running = False
    self.halted = False

    self.cpus = []
    self.memory = None

    self.devices = collections.defaultdict(dict)
    self.ports = {}

    self.virtual_interrupts = {}

    self.last_state = None

  @property
  def cores(self):
    """
    Get list of all cores in the machine.

    :rtype: list
    :returns: `list` of :py:class:`ducky.cpu.CPUCore` instances
    """

    return [c for c in itertools.chain(*[__cpu.cores for __cpu in self.cpus])]

  def on_core_alive(self, core):
    """
    Signal machine that one of CPU cores is now alive.
    """

    self.living_cores.append(core)

  def on_core_halted(self, core):
    """
    Signal machine that one of CPU cores is no longer alive.
    """

    self.living_cores.remove(core)

    if not self.living_cores:
      self.reactor.task_runnable(self.check_living_cores_task)

  def get_device_by_name(self, name, klass = None):
    """
    Get device by its name and class.

    :param string name: name of the device.
    :param string klass: if set, search only devices with this class.
    :rtype: :py:class:`ducky.devices.Device`
    :returns: found device
    :raises ducky.errors.InvalidResourceError: when no such device exists
    """

    self.DEBUG('get_device_by_name: name=%s, klass=%s', name, klass)

    for dev_klass, devs in iteritems(self.devices):
      if klass and dev_klass != klass:
        continue

      for dev_name, dev in iteritems(devs):
        if dev_name != name:
          continue

        return dev

    raise InvalidResourceError(F('No such device: name={name}, klass={klass}', name = name, klass = klass))

  def get_storage_by_id(self, sid):
    """
    Get storage by its id.

    :param int sid: id of storage caller is looking for.
    :rtype: :py:class:`ducky.devices.Device`
    :returns: found device.
    :raises ducky.errors.InvalidResourceError: when no such storage exists.
    """

    self.DEBUG('get_storage_by_id: id=%s', sid)
    self.DEBUG('storages: %s', self.devices['storage'])

    for name, dev in iteritems(self.devices['storage']):
      if dev.sid != sid:
        continue

      return dev

    raise InvalidResourceError(F('No such storage: sid={sid:d}', sid = sid))

  def save_state(self, parent):
    state = parent.add_child('machine', MachineState())

    state.nr_cpus = self.nr_cpus
    state.nr_cores = self.nr_cores

    for cpu in self.cpus:
      cpu.save_state(state)

    self.memory.save_state(state)

  def load_state(self, state):
    self.nr_cpus = state.nr_cpus
    self.nr_cores = state.nr_cores

    for __cpu in self.cpus:
      cpu_state = state.get_children().get('cpu{}'.format(__cpu.id))
      if cpu_state is None:
        self.WARN('State of CPU #%i not found!', __cpu.id)
        continue

      __cpu.load_state(cpu_state)

    self.memory.load_state(state.get_children()['memory'])

  def setup_devices(self):
    from .devices import get_driver_creator

    for section in self.config.iter_devices():
      _get, _getbool, _getint = self.config.create_getters(section)

      klass = _get('klass', None)
      driver = _get('driver', None)

      if not klass or not driver:
        self.ERROR('Unknown class or driver of device %s: klass=%s, driver=%s', section, klass, driver)
        continue

      if _getbool('enabled', True) is not True:
        self.DEBUG('Device %s disabled', section)
        continue

      dev = get_driver_creator(driver)(self, self.config, section)
      self.devices[klass][section] = dev

      if _get('master', None) is not None:
        dev.master = _get('master')

  def hw_setup(self, machine_config):
    self.config = machine_config

    self._tenh_enabled = machine_config.getbool('machine', 'tenh-enabled', False)

    self.nr_cpus = self.config.getint('machine', 'cpus')
    self.nr_cores = self.config.getint('machine', 'cores')

    # self.ivt_address = machine_config.getint('cpu', 'ivt-address', DEFAULT_IVT_ADDRESS)
    # self.pt_address = machine_config.getint('cpu', 'pt-address', DEFAULT_PT_ADDRESS)

    self.memory = mm.MemoryController(self, size = machine_config.getint('memory', 'size', 0x1000000))

    self.setup_devices()

    self.rom_loader = ROMLoader(self)

    from .cpu import CPU
    for cpuid in range(0, self.nr_cpus):
      self.cpus.append(CPU(self, cpuid, self.memory, cores = self.nr_cores))

    from .devices import VIRTUAL_INTERRUPTS
    for index, cls in iteritems(VIRTUAL_INTERRUPTS):
      self.virtual_interrupts[index] = cls(self)

  @property
  def exit_code(self):
    return max([c.exit_code for c in itertools.chain(*[__cpu.cores for __cpu in self.cpus])])

  def register_port(self, port, handler):
    self.DEBUG('Machine.register_port: port=%s, handler=%s', UINT16_FMT(port), handler)

    if port in self.ports:
      raise IndexError('Port already assigned: {}'.format(UINT16_FMT(port)))

    self.ports[port] = handler

  def unregister_port(self, port):
    self.DEBUG('Machine.unregister_port: port=%s', UINT16_FMT(port))

    del self.ports[port]

  def trigger_irq(self, handler):
    self.DEBUG('Machine.trigger_irq: handler=%s', handler)

    self.irq_router_task.queue[handler.irq] = True
    self.reactor.task_runnable(self.irq_router_task)

  def _do_tenh(self, printer, s, *args):
    printer('  ' + s + '\r\n', *args)
    self.INFO(s, *args)

  def tenh(self, s, *args):
    if not self._tenh_enabled:
      self.INFO(s, *args)
      return

    if self._tenh is None:
      for name, device in iteritems(self.devices['output']):
        if hasattr(device, 'tenh'):
          self._tenh_device = device
          self._tenh = partial(self._do_tenh, device.tenh)
          device.tenh_enable()
          break

      else:
        self._tenh = self.INFO

    self._tenh(s, *args)

  def boot(self):
    self.tenh('Ducky VM, version %s', __version__)
    self.tenh('Running on %s', sys.version.replace('\n', ' '))

    if self.config.getbool('machine', 'jit', False) is True:
      self.tenh('JIT enabled')

    self.DEBUG('Machine.boot')

    self.events.add_listener('on-core-alive', self.on_core_alive)
    self.events.add_listener('on-core-halted', self.on_core_halted)

    self.memory.boot()
    self.console.boot()

    for devs in itervalues(self.devices):
      for dev in [dev for dev in itervalues(devs) if not dev.is_slave()]:
        dev.boot()

    self.rom_loader.boot()

    for __cpu in self.cpus:
      __cpu.boot()

    self.running = True

  def run(self):
    self.DEBUG('Machine.run')

    for devs in itervalues(self.devices):
      for dev in [dev for dev in itervalues(devs) if not dev.is_slave()]:
        dev.run()

    for __cpu in self.cpus:
      __cpu.run()

    self.start_time = self.end_time = time.time()
    self.reactor.run()
    self.end_time = time.time()

  def suspend(self):
    self.DEBUG('Machine.suspend')

    for __cpu in self.cpus:
      __cpu.suspend()

  def wake_up(self):
    self.DEBUG('Machine.wake_up')

    for __cpu in self.cpus:
      __cpu.wake_up()

  def die(self, exc):
    self.DEBUG('Machine.die: exc=%s', exc)

    self.EXCEPTION(exc)

    self.halt()

  def halt(self):
    self.DEBUG('Machine.halt')

    self.capture_state()

    for __cpu in self.cpus:
      __cpu.halt()

    for devs in itervalues(self.devices):
      for dev in [dev for dev in itervalues(devs) if not dev.is_slave()]:
        dev.halt()

    self.rom_loader.halt()

    self.memory.halt()

    self.console.halt()

    self.reactor.remove_task(self.irq_router_task)
    self.reactor.remove_task(self.check_living_cores_task)

    self.events.remove_listener('on-core-alive', self.on_core_alive)
    self.events.remove_listener('on-core-halted', self.on_core_halted)

    self.tenh('Halted.')

    if self._tenh_enabled is True:
      self._tenh_device.tenh_flush_stream()
      self._tenh_device.tenh_close_stream()

    self.running = False
    self.halted = True

  def capture_state(self, suspend = False):
    """
    Capture current state of the VM, and store it in it's `last_state` attribute.

    :param bool suspend: if `True`, suspend VM before taking snapshot.
    """

    self.last_state = snapshot.VMState.capture_vm_state(self, suspend = suspend)
    return self.last_state

def cmd_boot(console, cmd):
  """
  Setup HW, load binaries, init everything
  """

  M = console.master.machine

  M.boot()
  M.console.unregister_command('boot')

def cmd_run(console, cmd):
  """
  Start execution of loaded binaries
  """

  M = console.master.machine

  M.run()
  M.console.unregister_command('run')

def cmd_halt(console, cmd):
  """
  Halt execution
  """

  M = console.master.machine

  M.halt()

  M.INFO('VM halted by user')

def cmd_snapshot(console, cmd):
  """
  Create snapshot
  """

  M = console.master.machine

  state = snapshot.VMState.capture_vm_state(M)

  filename = 'ducky-core.{}'.format(os.getpid())
  state.save(filename)

  M.INFO('Snapshot saved as %s', filename)
  console.writeln('Snapshot saved as %s', filename)



from . import BlockVisitor
from .. import Comment, J, MUL, ADD, LI, STB, LB

class BTPeepholeVisitor(BlockVisitor):
  priority = 100

  def do_visit_block(self, block):
    insts = block.instructions()
    for i in insts:
      if isinstance(i, J):
        index = insts.index(i)

        if index == 0:
          continue

        j = insts[index - 1]

        if not isinstance(j, J):
          continue

        self.DEBUG(self.log_prefix + 'Instruction "%s" follows previous jump instruction "%s", it is effectively dead', i, j)
        block.code[block.code.index(i)] = Comment('instruction "%s" removed by bt-peephole opt pass' % i)
        continue

      if isinstance(i, MUL) and i.operands[1] == 2:
        self.DEBUG(self.log_prefix + 'Instruction "%s" can be replaced by faster one', i)
        block.code[block.code.index(i)] = ADD(i.operands[0], i.operands[0])
        continue

      if isinstance(i, MUL) and i.operands[1] == 1:
        self.DEBUG(self.log_prefix + 'Instruction "%s" is needles, remove', i)
        block.code[block.code.index(i)] = Comment('instruction "%s" removed by bt-peephole opt pass' % i)
        continue

      if isinstance(i, MUL) and i.operands[1] == 0:
        self.DEBUG(self.log_prefix + 'Instruction "%s" can be replaced by faster one', i)
        block.code[block.code.index(i)] = LI(i.operands[0], 0)
        continue

      if isinstance(i, ADD) and i.operands[1] == 0:
        self.DEBUG(self.log_prefix + 'Instruction "%s" is needles, remove', i)
        block.code[block.code.index(i)] = Comment('instruction "%s" removed by bt-peephole opt pass' % i)
        continue

      if isinstance(i, STB):
        index = insts.index(i)

        if index == len(insts) - 1:
          continue

        j = insts[index + 1]

        if not isinstance(j, LB):
          continue

        if i.operands[0] != j.operands[1] or i.operands[1] != j.operands[0]:
          continue

        self.DEBUG(self.log_prefix + 'Instruction "%s" follows "%s", the later is needles. Removing.', j, i)
        block.code[block.code.index(j)] = Comment('instruction "%s" removed by bt-peephole opt pass' % j)
        continue

from . import BT_PASSES
BT_PASSES['bt-peephole'] = BTPeepholeVisitor



"""
Very simple character device that just "prints" characters on the screen.
It does not care about dimensions of the display, it kknow only how to
"print" characters. Suited for the most basic output possible - just "print"
chars by writing to this device, and you'll get this written into a stream
attached to the frontend (``stdout``, file, ...).
"""

from . import IOProvider, DeviceFrontend, DeviceBackend
from ..errors import InvalidResourceError
from ..mm import UINT16_FMT, UINT8_FMT
from ..interfaces import IReactorTask

DEFAULT_PORT_RANGE = 0x200

class FrontendFlushTask(IReactorTask):
  def __init__(self, frontend, queue, stream):
    super(FrontendFlushTask, self).__init__()

    self._machine = frontend.machine
    self._queue = queue
    self._stream = stream

  def set_output(self, stream):
    self._stream = stream

  def run(self):
    self._machine.DEBUG('%s.run', self.__class__.__name__)

    b = self._queue.read_out()
    if b is None:
      self._machine.DEBUG('%s.run: no events', self.__class__.__name__)
      return

    self._machine.DEBUG('%s.run: event=%r', self.__class__.__name__, b)
    self._stream.write([b])

class Frontend(DeviceFrontend):
  def __init__(self, machine, name):
    super(Frontend, self).__init__(machine, self.__class__, name)

    self._comm_queue = machine.comm_channel.get_queue(name)
    self._stream = None

    self._flush_task = None
    self.set_backend(machine.get_device_by_name(name))
    self.backend.set_frontend(self)

  @staticmethod
  def create_from_config(machine, config, section):
    slave = config.get(section, 'slave', default = section)

    return Frontend(machine, slave)

  def boot(self):
    super(Frontend, self).boot()

    self._flush_task = FrontendFlushTask(self, self._comm_queue, self._stream)
    self.machine.reactor.add_task(self._flush_task)
    self.machine.reactor.task_runnable(self._flush_task)

    self.backend.boot()

  def halt(self):
    self.backend.halt()

    self.machine.reactor.remove_task(self._flush_task)

    super(Frontend, self).halt()

  def set_output(self, stream):
    self.machine.DEBUG('%s.set_output: stream=%s', self.__class__.__name__, stream)

    self._stream = stream

    if self._flush_task is not None:
      self._flush_task.set_output(stream)

  def flush(self):
    while not self._comm_queue.is_empty_out():
      self._flush_task.run()

  def close(self, allow = False):
    if allow is True:
      self._stream.allow_close = True

    self._stream.close()

  def tenh_enable(self):
    if self._stream is not None:
      self._stream.allow_close = False

class Backend(IOProvider, DeviceBackend):
  def __init__(self, machine, name, stream = None, port = None, *args, **kwargs):
    super(Backend, self).__init__(machine, 'output', name, *args, **kwargs)

    self.port = port or DEFAULT_PORT_RANGE
    self.ports = [port]

    self.comm_queue = machine.comm_channel.create_queue(name)

  @staticmethod
  def create_from_config(machine, config, section):
    return Backend(machine, section,
                   port = config.getint(section, 'port', DEFAULT_PORT_RANGE))

  def __repr__(self):
    return 'basic tty on [%s] as %s' % (', '.join([UINT16_FMT(port) for port in self.ports]), self.name)

  def write_u8(self, port, value):
    self.machine.DEBUG('%s.write_u8: port=%s, value=%s', self.__class__.__name__, UINT16_FMT(port), UINT8_FMT(value))

    if port not in self.ports:
      raise InvalidResourceError('Unhandled port: %s' % UINT16_FMT(port))

    self.comm_queue.write_out(value)

  def tenh(self, s, *args):
    self.machine.DEBUG('%s.tenh: s="%s", args=%s', self.__class__.__name__, s, args)

    s = s % args

    for c in s:
      self.comm_queue.write_out(ord(c))

  def tenh_enable(self):
    self.frontend.tenh_enable()

  def tenh_flush_stream(self):
    self.frontend.flush()

  def tenh_close_stream(self):
    self.frontend.close(allow = True)

  def boot(self):
    self.machine.DEBUG('%s.boot', self.__class__.__name__)

    for port in self.ports:
      self.machine.register_port(port, self)

    self.machine.tenh('hid: %s', self)

  def halt(self):
    self.machine.DEBUG('%s.halt', self.__class__.__name__)

    for port in self.ports:
      self.machine.unregister_port(port)



import collections
import os
import sys

from six import iteritems

from ..mm import u32_t, UINT32_FMT, u16_t, u8_t, UINT8_FMT, MalformedBinaryError
from ..mm.binary import File, SectionTypes, SymbolEntry, SECTION_ITEM_SIZE, SectionFlags
from ..cpu.assemble import align_to_next_page, align_to_next_mmap, sizeof
from ..util import str2int
from ..errors import UnalignedJumpTargetError, EncodingLargeValueError, IncompatibleLinkerFlagsError, UnknownSymbolError

def align_nop(n):
  return n

class LinkerInfo(object):
  def __init__(self):
    super(LinkerInfo, self).__init__()

    self.section_offsets = collections.defaultdict(dict)
    self.relocations = collections.defaultdict(list)
    self.symbols = collections.defaultdict(list)
    self.section_bases = collections.defaultdict(dict)

def merge_object_into(logger, info, f_dst, f_src):
  r_header, r_content = f_src.get_section_by_name('.reloc')

  for s_header, s_content in f_src.sections():
    if s_header.type == SectionTypes.RELOC:
      info.relocations[f_src].append((s_header, s_content))
      continue

    elif s_header.type == SectionTypes.SYMBOLS:
      info.symbols[f_src].append((s_header, s_content))
      continue

    elif s_header.type == SectionTypes.STRINGS:
      continue

    s_name = f_src.string_table.get_string(s_header.name)

    logger.debug('Merge section %s into dst file', s_name)

    align = align_to_next_mmap if f_dst.get_header().flags.mmapable == 1 else align_nop

    try:
      d_header, d_content = f_dst.get_section_by_name(s_name)

    except MalformedBinaryError:
      logger.debug('No such section exists in dst file yet, copy')

      info.section_offsets[f_src][s_header.index] = 0
      info.section_bases[f_src][s_header.index] = s_header.base

      d_header = f_dst.create_section()
      d_header.type = s_header.type
      d_header.items = s_header.items
      d_header.data_size = s_header.data_size
      d_header.file_size = align(d_header.data_size)
      d_header.name = f_dst.string_table.put_string(s_name)
      d_header.flags = s_header.flags
      f_dst.set_content(d_header, s_content)
      continue

    if SectionFlags.from_encoding(d_header.flags).to_int() != SectionFlags.from_encoding(s_header.flags).to_int():
      logger.error('Source section has different flags set: d_header=%s, s_header=%s, f_src=%s', d_header, s_header, f_src)
      sys.exit(1)

    logger.debug('Merging into an existing section: %s', d_header)

    info.section_offsets[f_src][s_header.index] = d_header.data_size
    info.section_bases[f_src][s_header.index] = s_header.base

    d_header.items += s_header.items
    d_header.data_size += s_header.data_size
    d_header.file_size = align(d_header.data_size)
    d_content += s_content
    f_dst.set_content(d_header, d_content)

  f_dst.save()

def fix_section_bases(logger, info, f_out, required_bases):
  D = logger.debug

  D('Fixing base addresses of sections')

  sections_to_fix = {}

  for s_header, s_content in f_out.sections():
    s_header.base = 0xFFFFFFFF

    if s_header.type in (SectionTypes.RELOC, SectionTypes.SYMBOLS, SectionTypes.STRINGS):
      continue

    sections_to_fix[f_out.string_table.get_string(s_header.name)] = s_header

  D('sections to fix: %s', sections_to_fix)

  for s_name, s_base in iteritems(required_bases):
    if s_name not in sections_to_fix:
      continue

    s_header = sections_to_fix[s_name]

    s_header.base = s_base
    del sections_to_fix[s_name]

  D('sections to fix without required bases: %s', sections_to_fix)

  fixed_sections = []
  for s_header, s_content in f_out.sections():
    if s_header.type in (SectionTypes.RELOC, SectionTypes.SYMBOLS, SectionTypes.STRINGS):
      continue
    if s_header.base == 0xFFFFFFFF:
      continue

    fixed_sections.append(s_header)

  D('fixed sections: %s', fixed_sections)
  fixed_sections = sorted(fixed_sections, key = lambda x: x.base)
  D('fixed sections, sorted: %s', fixed_sections)

  align = align_to_next_mmap if f_out.get_header().flags.mmapable == 1 else align_to_next_page

  def overlap(a, b):
    return max(0, min(a[1], b[1]) - max(a[0], b[0]))

  for s_name, s_header in iteritems(sections_to_fix):
    base = 0

    for i, f_header in enumerate(fixed_sections):
      s_area = (base, align(base + s_header.data_size))
      f_area = (f_header.base, align(f_header.base + f_header.data_size))
      D('s_area: from=%s to=%s' % (UINT32_FMT(s_area[0]), UINT32_FMT(s_area[1])))
      D('f_area: from=%s to=%s' % (UINT32_FMT(f_area[0]), UINT32_FMT(f_area[1])))

      if overlap(s_area, f_area) != 0:
        base = f_area[1]
        D('  colides with f_area')
        continue

      if i < len(fixed_sections) - 1:
        f_next = fixed_sections[i + 1]
        if s_area[1] > f_next.base:
          base = f_next[1]
          D('   can not fit between f_header and f_next')
          continue

      D('  fits in')
      s_header.base = base
      fixed_sections.append(s_header)
      fixed_sections = sorted(fixed_sections, key = lambda x: x.base)
      break

    else:
      if fixed_sections:
        base = align(fixed_sections[-1].base + fixed_sections[-1].data_size)
      else:
        base = 0

      if base + s_header.data_size >= 0xFFFFFFFF:
        logger.error('Cant fit %s into any space', s_name)
        logger.error('section: %s', s_header)
        logger.error('fixed: %s', fixed_sections)
        sys.exit(1)

      D('  append at the end')
      s_header.base = base
      fixed_sections.append(s_header)
      fixed_sections = sorted(fixed_sections, key = lambda x: x.base)

  f_out.save()

def resolve_symbols(logger, info, f_out, f_ins):
  D = logger.debug

  D('Resolve symbols - compute their new addresses')

  symbols = []
  symbol_map = []

  for f_in, symbol_sections in iteritems(info.symbols):
    D('Processing file %s', f_in.name)

    for s_header, s_content in symbol_sections:
      D('Symbol section: %s', s_header)

      for s_se in s_content:
        s_name = f_in.string_table.get_string(s_se.name)

        D('Symbol: %s', s_name)

        o_header, o_content = f_in.get_section(s_se.section)
        D('Symbol points to source section: %s', o_header)

        o_name = f_in.string_table.get_string(o_header.name)

        d_header, d_content = f_out.get_section_by_name(o_name)
        D('Symbol points to destination section: %s', d_header)

        D('src base: %s, dst base: %s, symbol addr: %s, section dst offset: %s', UINT32_FMT(o_header.base), UINT32_FMT(d_header.base), UINT32_FMT(s_se.address), UINT32_FMT(info.section_offsets[f_in][o_header.index]))
        new_addr = s_se.address - o_header.base + info.section_offsets[f_in][o_header.index] + d_header.base

        d_se = SymbolEntry()
        d_se.flags = s_se.flags
        d_se.name = f_out.string_table.put_string(f_in.string_table.get_string(s_se.name))
        d_se.address = new_addr
        d_se.size = s_se.size
        d_se.section = d_header.index
        d_se.type = s_se.type
        d_se.filename = f_out.string_table.put_string(f_in.string_table.get_string(s_se.filename))
        d_se.lineno = s_se.lineno

        symbols.append(d_se)
        symbol_map.append((s_name, f_in, d_se))

        D('New symbol: %s', d_se)

  h_symtab = f_out.create_section()
  h_symtab.type = SectionTypes.SYMBOLS
  h_symtab.items = len(symbols)
  h_symtab.name = f_out.string_table.put_string('.symtab')
  h_symtab.base = 0xFFFFFFFF

  f_out.set_content(h_symtab, symbols)
  h_symtab.data_size = h_symtab.file_size = len(symbols) * sizeof(SymbolEntry())

  f_out.save()

  info.symbols = symbol_map

def resolve_relocations(logger, info, f_out, f_ins):
  D = logger.debug

  D('Resolve relocations')

  for s_header, s_content in f_out.sections():
    D('name=%s base=%s header=%s', f_out.string_table.get_string(s_header.name), UINT32_FMT(s_header.base), s_header)

  for f_in, reloc_sections in iteritems(info.relocations):
    D('Processing file %s', f_in.name)

    for r_header, r_content in reloc_sections:
      for re in r_content:
        D('-----*-----*-----')
        D('Relocation: reloc=%s', re)

        s_name = f_in.string_table.get_string(re.name)

        o_header, o_content = f_in.get_section(re.patch_section)
        D('Reloc points to source section: %s', o_header)

        o_name = f_in.string_table.get_string(o_header.name)

        d_header, d_content = f_out.get_section_by_name(o_name)
        D('Reloc points to destination section: %s', d_header)

        D('patch address: %s, patch offset: %s, patch size: %s, old base: %s, new base: %s', UINT32_FMT(re.patch_address), re.patch_offset, re.patch_size, o_header.base, d_header.base)
        fixed_patch_address = re.patch_address - o_header.base + info.section_offsets[f_in][o_header.index] + d_header.base

        D('fixed patch address: %s', UINT32_FMT(fixed_patch_address))

        D('Search for symbol named "%s"', s_name)

        for name, f_src, se in info.symbols:
          if name != s_name:
            continue

          if f_src != f_in and se.flags.globally_visible == 0:
            continue

          D('  Found: %s', se)
          break

        else:
          raise UnknownSymbolError('No such symbol: name=%s' % s_name)

        patch_address = se.address
        if re.flags.relative == 1:
          patch_address -= (fixed_patch_address + 4)

        D('Patching %s:%s:%s with %s', UINT32_FMT(fixed_patch_address), re.patch_offset, re.patch_size, UINT32_FMT(patch_address))

        content_index = (fixed_patch_address - d_header.base) // SECTION_ITEM_SIZE[d_header.type]
        D('  Content index: %i (%s - %s) / %s', content_index, fixed_patch_address, d_header.base, SECTION_ITEM_SIZE[d_header.type])

        orig_val = d_content[content_index]

        if isinstance(orig_val, u8_t):
          if re.patch_offset != 0 or re.patch_size != 16:
            logger.warn('Unhandled reloc entry: %s', re)
            sys.exit(1)

          bl = orig_val
          bh = d_content[content_index + 1]

          patch = u16_t(patch_address)
          new_bl = u8_t(patch.value & 0x00FF)
          new_bh = u8_t(patch.value >> 8)

          D('  patched! %s %s => %s %s', UINT8_FMT(bl), UINT8_FMT(bh), UINT8_FMT(new_bl), UINT8_FMT(new_bh))

          bl.value = new_bl.value
          bh.value = new_bh.value

        elif isinstance(orig_val, u32_t):
          new_val = None

          lower_mask = u32_t(0xFFFFFFFF >> (32 - re.patch_offset))
          upper_mask = u32_t(0xFFFFFFFF << (re.patch_offset + re.patch_size))
          mask = u32_t(upper_mask.value | lower_mask.value)
          D('lower mask: %s', UINT32_FMT(lower_mask))
          D('upper mask: %s', UINT32_FMT(upper_mask))
          D('mask: %s', UINT32_FMT(mask))

          D('orig val: %s', UINT32_FMT(orig_val))

          masked = u32_t(orig_val.value & mask.value)

          D('masked: %s', UINT32_FMT(masked.value))

          patch = patch_address + re.patch_add

          if re.flags.inst_aligned == 1:
            if patch & 0x3:
              raise UnalignedJumpTargetError(info = 'address=%s' % UINT32_FMT(patch))

            patch >>= 2

          if patch >= 2 ** re.patch_size:
            raise EncodingLargeValueError(info = 'size=%s, value=%s' % (re.patch_size, UINT32_FMT(patch)))

          patch = u32_t(patch << re.patch_offset)
          D('patch: %s', UINT32_FMT(patch))

          patch.value &= (~mask.value)
          D('patch: %s', UINT32_FMT(patch))

          new_val = u32_t(masked.value | patch.value)
          D('new val: %s', UINT32_FMT(new_val))

          if orig_val.value != new_val.value:
            D('  patched! %s => %s', UINT32_FMT(orig_val), UINT32_FMT(new_val))

          orig_val.value = new_val.value

        else:
          logger.warn('Unhandled content type: %s', orig_val)
          sys.exit(1)

  f_out.save()

def process_files(logger, info, files_in, file_out, bases = None):
  bases = bases or {}

  fs_in = []

  for file_in in files_in:
    with File.open(logger, file_in, 'r') as f_in:
      f_in.load()
      fs_in.append(f_in)

  if not all([f.get_header().flags.mmapable == fs_in[0].get_header().flags.mmapable for f in fs_in]):
    raise IncompatibleLinkerFlagsError()

  with File.open(logger, file_out, 'w') as f_out:
    h_file = f_out.create_header()
    h_file.flags.mmapable = fs_in[0].get_header().flags.mmapable

    h_section = f_out.create_section()
    h_section.type = SectionTypes.STRINGS
    h_section.name = f_out.string_table.put_string('.strings')

    for f_in in fs_in:
      merge_object_into(logger, info, f_out, f_in)

    fix_section_bases(logger, info, f_out, bases)
    resolve_symbols(logger, info, f_out, fs_in)
    resolve_relocations(logger, info, f_out, fs_in)

    f_out.save()

def main():
  import optparse
  from . import add_common_options, parse_options

  parser = optparse.OptionParser()
  add_common_options(parser)

  group = optparse.OptionGroup(parser, 'File options')
  parser.add_option_group(group)
  group.add_option('-i', dest = 'file_in', action = 'append', default = [], help = 'Input file')
  group.add_option('-o', dest = 'file_out', default = None, help = 'Output file')
  group.add_option('-f', dest = 'force', default = False, action = 'store_true', help = 'Force overwrite of the output file')

  group = optparse.OptionGroup(parser, 'Linker options')
  parser.add_option_group(group)
  group.add_option('--section-base', dest = 'section_base', action = 'append', default = [], help = 'Set base of section to specific address', metavar = 'SECTION=ADDRESS')

  options, logger = parse_options(parser)

  if not options.file_in:
    parser.print_help()
    sys.exit(1)

  if not options.file_out:
    parser.print_help()
    sys.exit(1)

  if any(not file_in.endswith('.o') for file_in in options.file_in):
    logger.error('All input files must be object files')
    sys.exit(1)

  if os.path.exists(options.file_out) and not options.force:
    logger.error('Output file %s already exists, use -f to force overwrite', options.file_out)
    sys.exit(1)

  info = LinkerInfo()

  try:
    process_files(logger, info, options.file_in, options.file_out, bases = {name: str2int(value) for name, value in (e.split('=') for e in options.section_base)})

  except IncompatibleLinkerFlagsError:
    logger.error('All input files must have the same mmapable setting')
    os.unlink(options.file_out)
    sys.exit(1)

  except UnknownSymbolError as e:
    logger.exception(e)
    os.unlink(options.file_out)
    sys.exit(1)



"""
This file provides necessary code to allow boot up of a virtual machine with
the correct program running. This code can provide slightly different environment
when compared to real hardware process, since e.g. external files can be mmap-ed
into VM's memory for writing.
"""

import importlib
import mmap

from functools import partial
from ctypes import sizeof
from six import PY2

from .interfaces import IMachineWorker

from .errors import InvalidResourceError
from .util import align, BinaryFile
from .mm import u8_t, u16_t, u32_t, UINT32_FMT, PAGE_SIZE, area_to_pages, PAGE_MASK, ExternalMemoryPage
from .mm.binary import SectionFlags, File, SectionTypes
from .snapshot import SnapshotNode
from .hdt import HDT, HDTEntry_Argument
from .debugging import Point  # noqa

#: By default, Hardware Description Table starts at this address after boot.
DEFAULT_HDT_ADDRESS        = 0x00000100

#: By default, CPU starts executing instructions at this address after boot.
DEFAULT_BOOTLOADER_ADDRESS = 0x00020000

class MMapMemoryPage(ExternalMemoryPage):
  """
  Memory page backed by an external file that is accessible via ``mmap()``
  call. It's a part of one of mm.MMapArea instances, and if such area was
  opened as `shared`, every change in this page content will affect the
  content of external file, and vice versa, every change of external file
  will be reflected in content of this page (if this page lies in affected
  area).

  :param MMapArea area: area this page belongs to.
  """

  def __init__(self, area, *args, **kwargs):
    super(MMapMemoryPage, self).__init__(*args, **kwargs)

    self.area = area

    if PY2:
      self.get, self.put = self._get_py2, self._put_py2

    else:
      self.get, self.put = self._get_py3, self._put_py3

  def get(self, offset):
    """
    Read one byte from page.

    This is an abstract method, `__init__` is expected to replace it with
    a method, tailored for the Python version used.

    :param int offset: offset of the requested byte.
    :rtype: int
    """

    raise NotImplementedError()

  def put(self, offset, b):
    """
    Write one byte to page.

    This is an abstract method, `__init__` is expected to replace it with
    a method, tailored for the Python version used.

    :param int offset: offset of the modified byte.
    :param int b: new value of the modified byte.
    """

    raise NotImplementedError()

  def _get_py2(self, offset):
    """
    Read one byte from page.

    :param int offset: offset of the requested byte.
    :rtype: int
    """

    return ord(self.data[self.offset + offset])

  def _put_py2(self, offset, b):
    """
    Write one byte to page.

    :param int offset: offset of the modified byte.
    :param int b: new value of the modified byte.
    """

    self.data[self.offset + offset] = chr(b)

  def _get_py3(self, offset):
    """
    Read one byte from page.

    :param int offset: offset of the requested byte.
    :rtype: int
    """

    return self.data[self.offset + offset]

  def _put_py3(self, offset, b):
    """
    Write one byte to page.

    :param int offset: offset of the modified byte.
    :param int b: new value of the modified byte.
    """

    self.data[self.offset + offset] = b

class MMapAreaState(SnapshotNode):
  def __init__(self):
    super(MMapAreaState, self).__init__('address', 'size', 'path', 'offset')

class MMapArea(object):
  """
  Objects of this class represent one mmaped memory area each, to track this
  information for later use.

  :param ptr: ``mmap object``, as returned by `mmap.mmap` function.
  :param u32_t address: address of the first byte of an area in the memory.
  :param u32_t size: length of the area, in bytes.
  :param file_path: path to a source file.
  :param u32_t offset: offset of the first byte in the source file.
  :param int pages_start: first page of the area.
  :param int pages_cnt: number of pages in the area.
  :param mm.binary.SectionFlags flags: flags applied to this area.
  """

  def __init__(self, ptr, address, size, file_path, offset, pages_start, pages_cnt, flags):
    super(MMapArea, self).__init__()

    self.ptr = ptr
    self.address = address

    self.size = size
    self.file_path = file_path
    self.offset = offset
    self.pages_start = pages_start
    self.pages_cnt = pages_cnt
    self.flags = flags

  def __repr__(self):
    return '<MMapArea: address=%s, size=%s, filepath=%s, pages-start=%s, pages-cnt=%i, flags=%s>' % (UINT32_FMT(self.address), self.size, self.file_path, self.pages_start, self.pages_cnt, self.flags.to_string())

  def save_state(self, parent):
    pass

  def load_state(self, state):
    pass

class ROMLoader(IMachineWorker):
  def __init__(self, machine):
    self.machine = machine
    self.config = machine.config

    self.opened_mmap_files = {}  # path: (cnt, file)
    self.mmap_areas = {}

    self.logger = self.machine.LOGGER
    self.DEBUG = self.machine.DEBUG

  def _get_mmap_fileno(self, file_path):
    if file_path not in self.opened_mmap_files:
      self.opened_mmap_files[file_path] = [0, open(file_path, 'r+b')]

    desc = self.opened_mmap_files[file_path]

    desc[0] += 1
    return desc[1].fileno()

  def _put_mmap_fileno(self, file_path):
    desc = self.opened_mmap_files[file_path]

    desc[0] -= 1
    if desc[0] > 0:
      return

    desc[1].close()
    del self.opened_mmap_files[file_path]

  def mmap_area(self, file_path, address, size, offset = 0, flags = None, shared = False):
    """
    Assign set of memory pages to mirror external file, mapped into memory.

    :param string file_path: path of external file, whose content new area
      should reflect.
    :param u24 address: address where new area should start.
    :param u24 size: length of area, in bytes.
    :param int offset: starting point of the area in mmaped file.
    :param ducky.mm.binary.SectionFlags flags: specifies required flags for mmaped
      pages.
    :param bool shared: if ``True``, content of external file is mmaped as
      shared, i.e. all changes are visible to all processes, not only to the
      current ducky virtual machine.
    :returns: newly created mmap area.
    :rtype: ducky.mm.MMapArea
    :raises ducky.errors.InvalidResourceError: when ``size`` is not multiply of
      :py:data:`ducky.mm.PAGE_SIZE`, or when ``address`` is not multiply of
      :py:data:`ducky.mm.PAGE_SIZE`, or when any of pages in the affected area
      is already allocated.
    """

    self.DEBUG('%s.mmap_area: file=%s, offset=%s, size=%s, address=%s, flags=%s, shared=%s', self.__class__.__name__, file_path, offset, size, UINT32_FMT(address), flags.to_string(), shared)

    if size % PAGE_SIZE != 0:
      raise InvalidResourceError('Memory size must be multiple of PAGE_SIZE')

    if address % PAGE_SIZE != 0:
      raise InvalidResourceError('MMap area address must be multiple of PAGE_SIZE')

    mc = self.machine.memory
    pages_start, pages_cnt = area_to_pages(address, size)

    for i in range(pages_start, pages_start + pages_cnt):
      if i in mc.pages:
        raise InvalidResourceError('MMap request overlaps with existing pages: page=%s, area=%s' % (mc.pages[i], mc.pages[i].area))

    mmap_flags = mmap.MAP_SHARED if shared else mmap.MAP_PRIVATE

    # Always mmap as writable - VM will force read-only access using
    # page flags. But since it is possible to change page flags
    # in run-time, and request write access to areas originaly
    # loaded as read-only, such write access would fail because
    # the underlying mmap area was mmaped as read-only only, and this
    # limitation is not possible to overcome.
    mmap_prot = mmap.PROT_READ | mmap.PROT_WRITE

    ptr = mmap.mmap(
      self._get_mmap_fileno(file_path),
      size,
      flags = mmap_flags,
      prot = mmap_prot,
      offset = offset)

    area = MMapArea(ptr, address, size, file_path, ptr, pages_start, pages_cnt, flags)

    for i in range(pages_start, pages_start + pages_cnt):
      mc.register_page(MMapMemoryPage(area, mc, i, ptr, offset = (i - pages_start) * PAGE_SIZE))

    self.mmap_areas[area.address] = area

    return area

  def unmmap_area(self, mmap_area):
    mc = self.machine.memory

    for pg in mc.get_pages(pages_start = mmap_area.pages_start, pages_cnt = mmap_area.pages_cnt):
      mc.unregister_page(pg)

    del self.mmap_areas[mmap_area.address]

    mmap_area.ptr.close()

    self._put_mmap_fileno(mmap_area.file_path)

  def setup_hdt(self):
    """
    Initialize memory area that contains HDT.

    If VM config file specifies HDT image file, it is loaded, otherwise HDT
    is constructed for the actual configuration. It is then copied into memory.

    :param u32_t machine.hdt-address: Base address of HDT in memory. If not set,
      :py:const:`ducky.boot.DEFAULT_HDT_ADDRESS` is used.
    :param machine.hdt-image: HDT image to load. If not set, HDT is constructed
      for the actual VM's configuration.
    """

    self.DEBUG('%s.setup_hdt', self.__class__.__name__)

    hdt_address = self.config.getint('machine', 'hdt-address', DEFAULT_HDT_ADDRESS)
    if hdt_address & ~PAGE_MASK:
      raise InvalidResourceError('HDT address must be page-aligned: address=%s' % UINT32_FMT(hdt_address))

    self.DEBUG('HDT address=%s', UINT32_FMT(hdt_address))

    def __alloc_pages(size):
      pages = self.machine.memory.alloc_pages(base = hdt_address, count = align(PAGE_SIZE, size) // PAGE_SIZE)
      self.machine.DEBUG('%s.setup_hdt: address=%s, size=%s (%s pages)', self.__class__.__name__, UINT32_FMT(hdt_address), size, len(pages))

    hdt_image = self.config.get('machine', 'hdt-image', None)
    if hdt_image is None:
      self.DEBUG('HDT image not specified, creating one')

      hdt = HDT(self.machine.LOGGER, config = self.config)
      hdt.create()

      __alloc_pages(len(hdt))

      def __write_field(writer_fn, size, address, field_value):
        writer_fn(address, field_value)
        return address + size

      def __write_array(max_length, address, field_value):
        for i in range(0, max_length):
          self.machine.memory.write_u8(address + i, field_value[i])

        return address + max_length

      def __write_struct(address, struct):
        for n, t in struct._fields_:
          address = writers[sizeof(t)](address, getattr(struct, n))

        return address

      writers = {
        1: partial(__write_field, self.machine.memory.write_u8,  1),
        2: partial(__write_field, self.machine.memory.write_u16, 2),
        4: partial(__write_field, self.machine.memory.write_u32, 4),
        HDTEntry_Argument.MAX_NAME_LENGTH: partial(__write_array, HDTEntry_Argument.MAX_NAME_LENGTH)
      }

      address = __write_struct(hdt_address, hdt.header)

      for entry in hdt.entries:
        address = __write_struct(address, entry)

    else:
      self.DEBUG('Loading HDT image %s', hdt_image)

      with BinaryFile.open(self.logger, hdt_image, 'r') as f_in:
        img = f_in.read()

      __alloc_pages(len(img))

      for address, b in zip(range(hdt_address, hdt_address + len(img)), img):
        self.machine.memory.write_u8(address, b)

  def setup_mmaps(self):
    self.DEBUG('%s.setup_mmaps', self.__class__.__name__)

    for section in self.config.iter_mmaps():
      _get, _getbool, _getint = self.config.create_getters(section)

      access = _get('access', 'r')
      flags = SectionFlags.create(readable = 'r' in access, writable = 'w' in access, executable = 'x' in access)
      self.mmap_area(_get('file'), _getint('address'), _getint('size'), offset = _getint('offset', 0), flags = flags, shared = _getbool('shared', False))

  def setup_debugging(self):
    self.DEBUG('%s.setup_debugging', self.__class__.__name__)

    for section in self.config.iter_breakpoints():
      _get, _getint, _getbool = self.config.create_getters(section)

      core = self.machine.core(_get('core', '#0:#0'))
      core.init_debug_set()

      klass = _get('klass', 'ducky.debugging.BreakPoint').split('.')
      klass = getattr(importlib.import_module('.'.join(klass[0:-1])), klass[-1])

      p = klass.create_from_config(core.debug, self.config, section)
      core.debug.add_point(p, _get('chain', 'pre-step'))

      for action_section in _get('actions', '').split(','):
        action_section = action_section.strip()
        if not action_section:
          continue

        klass = self.config.get(action_section, 'klass').split('.')
        klass = getattr(importlib.import_module('.'.join(klass[0:-1])), klass[-1])

        a = klass.create_from_config(core.debug, self.config, action_section)
        p.actions.append(a)

  def __load_content(self, base, content):
    from .cpu.assemble import SpaceSlot

    ptr = base

    self.DEBUG('%s.__load_content: base=%s, items=%s', self.__class__.__name__, UINT32_FMT(base), len(content))

    def __write(writer, size, value):
      writer(ptr, value)
      return ptr + size

    writers = {
      1: partial(__write, self.machine.memory.write_u8,  1),
      2: partial(__write, self.machine.memory.write_u16, 2),
      4: partial(__write, self.machine.memory.write_u32, 4)
    }

    for i in content:
      self.DEBUG('%s.__load_content: ptr=%s, i=%s', self.__class__.__name__, UINT32_FMT(ptr), UINT32_FMT(i))

      if isinstance(i, SpaceSlot):
        ptr += i.size

      else:
        ptr = writers[sizeof(i)](i.value)

  def load_text(self, base, content):
    self.DEBUG('%s.load_text: base=%s', self.__class__.__name__, UINT32_FMT(base))

    self.__load_content(base, content)

  def load_data(self, base, content):
    self.DEBUG('%s.load_data: base=%s', self.__class__.__name__, UINT32_FMT(base))

    self.__load_content(base, content)

  def setup_bootloader(self, filepath, base = None):
    self.DEBUG('%s.setup_bootloader: filepath=%s, base=%s', self.__class__.__name__, filepath, UINT32_FMT(base) if base is not None else '<none>')

    base = base or DEFAULT_BOOTLOADER_ADDRESS
    mc = self.machine.memory

    with File.open(self.machine.LOGGER, filepath, 'r') as f:
      f.load()

      f_header = f.get_header()

      for i in range(0, f_header.sections):
        s_header, s_content = f.get_section(i)

        s_base = s_header.base

        self.DEBUG('%s.setup_bootloader: section=%s, base=%s', self.__class__.__name__, f.string_table.get_string(s_header.name), UINT32_FMT(s_base))

        if s_header.type == SectionTypes.SYMBOLS:
          continue

        if s_header.flags.loadable != 1:
          continue

        s_base = base + s_base

        pages_start, pages_cnt = area_to_pages(s_base, s_header.file_size)

        if f_header.flags.mmapable == 1:
          # Always mmap sections as RW, and disable W if section' flags requires that
          # Otherwise, when program asks Vm to enable W, any access would fail because
          # the underlying mmap area was not mmaped as writable
          self.mmap_area(f.name, s_base, s_header.file_size, offset = s_header.offset, flags = SectionFlags.from_encoding(s_header.flags), shared = False)

        else:
          for i in range(pages_start, pages_start + pages_cnt):
            mc.alloc_specific_page(i)

          if s_header.type == SectionTypes.TEXT:
            self.load_text(s_base, s_content)

          elif s_header.type == SectionTypes.DATA and s_header.flags.bss != 1:
            self.load_data(s_base, s_content)

  def poke(self, address, value, length):
    self.DEBUG('%s.poke: addr=%s, value=%s, length=%s', self.__class__.__name__, UINT32_FMT(address), UINT32_FMT(value), length)

    if length == 1:
      self.machine.memory.write_u8(address, u8_t(value).value)

    elif length == 2:
      self.machine.memory.write_u16(address, u16_t(value).value)

    else:
      self.machine.memory.write_u32(address, u32_t(value).value)

  def boot(self):
    self.DEBUG('%s.boot', self.__class__.__name__)

    self.setup_hdt()
    self.setup_mmaps()
    self.setup_debugging()

    if self.config.has_section('bootloader'):
      self.setup_bootloader(self.config.get('bootloader', 'file'))

  def halt(self):
    self.DEBUG('%s.halt', self.__class__.__name__)

    for area in list(self.mmap_areas.values()):
      self.unmmap_area(area)
